class CoverageData:
    def __init__(self, coverage_type: str, executed_lines: int, missed_lines: int, branch_coverage: float, source_file_id: int, function_id: int):
        self._coverage_type: str = coverage_type
        self._executed_lines: int = executed_lines
        self._missed_lines: int = missed_lines
        self._branch_coverage: float = branch_coverage
        self._source_file_id: int = source_file_id
        self._function_id: int = function_id

    @property
    def coverage_type(self) -> str:
        return self._coverage_type

    @coverage_type.setter
    def coverage_type(self, value: str) -> None:
        self._coverage_type = value

    @property
    def executed_lines(self) -> int:
        return self._executed_lines

    @executed_lines.setter
    def executed_lines(self, value: int) -> None:
        self._executed_lines = value

    @property
    def missed_lines(self) -> int:
        return self._missed_lines

    @missed_lines.setter
    def missed_lines(self, value: int) -> None:
        self._missed_lines = value

    @property
    def branch_coverage(self) -> float:
        return self._branch_coverage

    @branch_coverage.setter
    def branch_coverage(self, value: float) -> None:
        self._branch_coverage = value

    @property
    def source_file_id(self) -> int:
        return self._source_file_id

    @source_file_id.setter
    def source_file_id(self, value: int) -> None:
        self._source_file_id = value

    @property
    def function_id(self) -> int:
        return self._function_id

    @function_id.setter
    def function_id(self, value: int) -> None:
        self._function_id = value


from testgen.generator.test_generator import TestGenerator
from testgen.models.generator_context import GeneratorContext

AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3

class UnitTestGenerator(TestGenerator):
    def __init__(self, generator_context: GeneratorContext):
        super().__init__(generator_context)
        self.test_code = []

    def generate_test_header(self):

        self.test_code.append("import unittest\n")
        if self._generator_context.class_name == "" or self._generator_context.class_name is None:
            self.test_code.append(f"import {self._generator_context.import_path} as {self._generator_context.module.__name__}\n")
        else:
            self.test_code.append(f"from {self._generator_context.import_path} import {self._generator_context.class_name}\n")
        self.test_code.append(f"class Test{self._generator_context.class_name}(unittest.TestCase):  \n")

    def generate_test_function(self, unique_func_name, func_name, cases):
        print(f"Cases: {cases}")
        self.test_code.append(f"   def test_{unique_func_name}(self):")
        for inputs, expected in cases:
            args_str: str = self.generate_args_statement(inputs)
            expected_str: str = self.generate_expected_statement(expected)
            is_class: bool = self._generator_context.class_name != "" and self._generator_context.class_name is not None
            class_or_mod_name: str = self.generate_class_or_mod_name(is_class)
            result_statement: str = self.generate_result_statement(class_or_mod_name, func_name, inputs)
            assert_statement: str = self.generate_assert_statement()


            self.test_code.append(args_str)
            self.test_code.append(expected_str)
            self.test_code.append(result_statement)
            self.test_code.append(assert_statement)

    def save_file(self):
        with open(self._generator_context.output_path, "w") as f:
            f.write("\n".join(self.test_code))

    def generate_class_or_mod_name(self, is_class: bool) -> str:
        if is_class:
            return f"{self._generator_context.class_name}()"
        else:
            return f"{self._generator_context.module.__name__}"

    @staticmethod
    def generate_args_statement(inputs) -> str:
        input_args = ', '.join(map(repr, inputs))
        if len(inputs) == 1:
            return f"      args = {input_args}"
        else:
            return f"      args = ({input_args})"

    @staticmethod
    def generate_expected_statement(expected) -> str:
        if isinstance(expected, str):
            return f"      expected = '{expected}'"
        else:
            return f"      expected = {expected}"
        

    @staticmethod
    def generate_result_statement(class_or_mod_name, func_name, inputs) -> str:
        if len(inputs) == 1:
            return f"      result = {class_or_mod_name}.{func_name}(args)"
        else:
            return f"      result = {class_or_mod_name}.{func_name}(*args)"

    @staticmethod
    def generate_assert_statement() -> str:
        return f"      self.assertEqual(result, expected)\n"




import ast
from typing import List

import z3

import testgen.util.file_utils
from testgen.models.test_case import TestCase
from testgen.util.z3_utils import ast_to_z3
from testgen.util.z3_utils.constraint_extractor import extract_branch_conditions


def solve_branch_condition(file_name: str, class_name: str | None, func_node: ast.FunctionDef, uncovered_lines: List[int]) -> List[TestCase]:
    branch_conditions, param_types = extract_branch_conditions(func_node)
    uncovered_conditions = [bc for bc in branch_conditions if bc.line_number in uncovered_lines]
    test_cases = []

    for branch_condition in uncovered_conditions:
        z3_expr, z3_vars = ast_to_z3.ast_to_z3_constraint(branch_condition, param_types)
        solver = z3.Solver()
        solver.add(z3_expr)

        if solver.check() == z3.sat:
            model = solver.model()

            # Create default values for all parameters
            param_values = {}
            for param_name in param_types:
                # Skip 'self' parameter for class methods
                if param_name == 'self':
                    continue
                    
                # Set default values based on type
                if param_types[param_name] == "int":
                    param_values[param_name] = 0
                elif param_types[param_name] == "float":
                    param_values[param_name] = 0.0
                elif param_types[param_name] == "bool":
                    param_values[param_name] = False
                elif param_types[param_name] == "str":
                    param_values[param_name] = ""
                else:
                    param_values[param_name] = None

            # Update with model values where available
            for var_name, z3_var in z3_vars.items():
                if var_name in param_types:
                    try:

                        model_value = model.evaluate(z3_var)

                        if param_types[var_name] == "int":
                            param_values[var_name] = model_value.as_long()
                        elif param_types[var_name] == "float":
                            param_values[var_name] = float(model_value.as_decimal(10))
                        elif param_types[var_name] == "bool":
                            param_values[var_name] = z3.is_true(model_value)
                        elif param_types[var_name] == "str":
                            str_val = str(model_value)
                            if str_val.startswith('"') and str_val.endswith('"'):
                                str_val = str_val[1:-1]
                            param_values[var_name] = str_val
                        else:
                            param_values[var_name] = model[z3_var].as_long()
                    except Exception as e:
                        print(f"Couldn't get {var_name} from model: {e}")
                        # Keep the default value we already set

            # Ensure all parameters are included in correct order
            ordered_params = []
            for arg in func_node.args.args:
                arg_name = arg.arg
                if arg_name == 'self':  # Skip self parameter
                    continue
                if arg_name in param_values:
                    ordered_params.append(param_values[arg_name])
                else:
                    # Default value handling if parameter not in solution
                    if hasattr(arg, 'annotation') and arg.annotation:
                        if isinstance(arg.annotation, ast.Name):
                            if arg.annotation.id == 'int':
                                ordered_params.append(0)
                            elif arg.annotation.id == 'float':
                                ordered_params.append(0.0)
                            elif arg.annotation.id == 'bool':
                                ordered_params.append(False)
                            elif arg.annotation.id == 'str':
                                ordered_params.append('')
                            else:
                                ordered_params.append(None)
                        else:
                            ordered_params.append(None)
                    else:
                        ordered_params.append(None)

            func_name = func_node.name
            try:
                module = testgen.util.file_utils.load_module(file_name)
                if class_name is not None:
                    class_obj = getattr(module, class_name)
                    instance = class_obj()
                    func = getattr(instance, func_name)
                else:
                    func = getattr(module, func_name)
                result = func(*ordered_params)
                test_cases.append(TestCase(func_name, tuple(ordered_params), result))
            except Exception as e:
                print(f"Error executing function with Z3 solution for {func_name}: {e}")

    return test_cases


import sqlite3

def create_database(db_name="testgen.db"):
    conn = sqlite3.connect(db_name)
    cursor = conn.cursor()

    # Enable foreign key constraints
    cursor.execute("PRAGMA foreign_keys = ON;")

    # Create tables
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS TestSuite (
            id INTEGER PRIMARY KEY,
            name TEXT,
            creation_date TIMESTAMP
        );
    """)

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS SourceFile (
            id INTEGER PRIMARY KEY,
            path TEXT,
            lines_of_code INTEGER,
            last_modified TIMESTAMP
        );
    """)

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS Function (
            id INTEGER PRIMARY KEY,
            name TEXT,
            params TEXT,
            start_line INTEGER,
            end_line INTEGER,
            num_lines INTEGER,
            source_file_id INTEGER,
            UNIQUE(name, params, source_file_id),
            FOREIGN KEY (source_file_id) REFERENCES SourceFile(id)
        );
    """)

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS TestCase (
            id INTEGER PRIMARY KEY,
            expected_output TEXT,
            input TEXT,
            test_function TEXT,
            last_run_time TIMESTAMP,
            test_method_type INTEGER,
            test_suite_id INTEGER,
            function_id INTEGER,
            FOREIGN KEY (test_suite_id) REFERENCES TestSuite(id),
            FOREIGN KEY (function_id) REFERENCES Function(id),
            UNIQUE(function_id, input, expected_output)
        );
    """)

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS TestResult (
            id INTEGER PRIMARY KEY,
            test_case_id INTEGER,
            status BOOLEAN,
            error TEXT,
            execution_time TIMESTAMP,
            FOREIGN KEY (test_case_id) REFERENCES TestCase(id)
        );
    """)

    cursor.execute("""
        CREATE TABLE IF NOT EXISTS CoverageData (
            id INTEGER PRIMARY KEY,
            coverage_type TEXT CHECK(coverage_type IN ('file', 'function')),
            executed_lines INTEGER,
            missed_lines INTEGER,
            branch_coverage REAL,
            source_file_id INTEGER,
            function_id INTEGER,
            FOREIGN KEY (source_file_id) REFERENCES SourceFile(id),
            FOREIGN KEY (function_id) REFERENCES Function(id),
            CHECK (
                -- Only one of source_file_id or function_id is required based on coverage_type
                (coverage_type = 'file' AND source_file_id IS NOT NULL AND function_id IS NULL) OR
                (coverage_type = 'function' AND function_id IS NOT NULL)
            )
        );
    """)

    conn.commit()
    conn.close()
    print(f"Database '{db_name}' created successfully with all tables.")

if __name__ == "__main__":
    create_database()



class SourceFile:
    def __init__(self, path: str, lines_of_code: int, last_modified):
        self._path = path
        self._lines_of_code = lines_of_code
        self._last_modified = last_modified

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    @property
    def lines_of_code(self) -> int:
        return self._lines_of_code

    @lines_of_code.setter
    def lines_of_code(self, value: int) -> None:
        self._lines_of_code = value

    @property
    def last_modified(self):
        return self._last_modified

    @last_modified.setter
    def last_modified(self, value) -> None:
        self._last_modified = value


import inspect
from typing import Optional, List, Set, Tuple

import astor
from testgen.generator.test_generator import TestGenerator
from testgen.models.generator_context import GeneratorContext

class DocTestGenerator(TestGenerator):
    def __init__(self, generator_context: GeneratorContext):
        super().__init__(generator_context)

    def generate_test_header(self):
        pass
    
    def generate_test_function(self, unique_func_name, func_name, cases) -> None:
        """Generate a test function with doctests."""
        doctest_examples = []

        for inputs, expected in cases:
            input_args = ', '.join(map(repr, inputs))

            if self._generator_context.class_name:
                test_call = f"{self._generator_context.class_name}().{func_name}({input_args})"
            else:
                test_call = f"{func_name}({input_args})"

            result_str = repr(expected)
            # Format with >>> and proper spacing
            doctest = f">>> {test_call}\n{result_str}"
            doctest_examples.append(doctest)

        # Inject the doctests into the original function's docstring
        self._inject_docstring(func_name, doctest_examples)
        
    def save_file(self):
            """Save the generated test content to a file."""
            return self._generator_context.filepath
        
    def _inject_docstring(self, func_name, doctest_examples):
        """Inject doctest examples into the function's docstring."""
        module_path = self._generator_context.filepath
        
        # Read the source file
        with open(module_path, 'r') as f:
            lines = f.readlines()

        func_line_num, func_indent = self._get_func_line_num_and_indent(lines, func_name)
        
        if func_line_num == -1:
            print(f"Function {func_name} not found in {module_path}")
            return
        
        has_docstring, docstring_end = self._get_docstring_end(func_line_num, lines)

        all_examples, seen_examples = self._collect_existing_docstrings(func_line_num, lines, has_docstring, docstring_end)
        
        
        all_examples, seen_examples = self._add_new_docstrings(doctest_examples, all_examples, seen_examples)
        
        docstring_lines = self._create_docstrings(func_indent, all_examples)
        
        # Modify the source code
        if has_docstring:
            # Replace existing docstring
            lines = lines[:func_line_num + 1] + [line + '\n' for line in docstring_lines] + lines[docstring_end + 1:]
        else:
            # Insert new docstring after function definition
            lines = lines[:func_line_num + 1] + [line + '\n' for line in docstring_lines] + lines[func_line_num + 1:]
        
        # Write back to the file
        with open(module_path, 'w') as f:
            f.writelines(lines)

    def _extract_doctest_examples(self, docstring_lines: List[str]) -> List[Tuple[str, str]]:
        examples = []
        current_call = None
        current_result_lines = []
        
        # Process docstring line by line
        for line in docstring_lines:
            stripped_line = line.strip()
            
            # Check if this is a test call line (starts with >>>)
            if stripped_line.startswith('>>>'):
                # If we were already collecting a test, save it
                if current_call is not None:
                    result = '\n'.join([line.strip() for line in current_result_lines if line.strip()])
                    examples.append((current_call, result))
                    current_result_lines = []
                
                # Extract the call part (everything after >>>)
                current_call = stripped_line[3:].strip()
            
            # Otherwise, if we have a current call, add to the result
            elif current_call is not None:
                current_result_lines.append(stripped_line)
        
        # Don't forget the last example if there is one
        if current_call is not None and current_result_lines:
            result = '\n'.join([line.strip() for line in current_result_lines if line.strip()])
            examples.append((current_call, result))
        
        return examples
    
    
    def _is_class_method(self, function_name: str) -> bool:
        """Check if a function is a method of a class in the module."""
        for name, obj in inspect.getmembers(self._generator_context.module):
            if inspect.isclass(obj) and hasattr(obj, function_name):
                return True
        return False
    
    def _get_class_name(self, method_name: str) -> Optional[str]:
        """Get the name of the class that contains the method."""
        for name, obj in inspect.getmembers(self._generator_context.module):
            if inspect.isclass(obj) and hasattr(obj, method_name):
                return name
        return None
    
    def _get_func_line_num_and_indent(self, lines: List[str], func_name: str) -> Tuple[int, str]:
        func_line_num = -1
        func_indent = ""
        
        for i, line in enumerate(lines):
            line_stripped = line.lstrip()
            if line_stripped.startswith('def '):
                # Extract function name
                potential_name = line_stripped[4:].split('(')[0].strip()
                if potential_name == func_name:
                    func_line_num = i
                    # Calculate indentation by counting spaces before 'def'
                    func_indent = line[:len(line) - len(line_stripped)]
                    break
        return (func_line_num, func_indent)
    
    def _get_docstring_end(self, func_line_num: int, lines: List[str]) -> Tuple[bool, int]:
        has_docstring = False
        docstring_end = -1
        
        if func_line_num + 1 < len(lines) and '"""' in lines[func_line_num + 1]:
            has_docstring = True
            # Find the end of the docstring
            for i in range(func_line_num + 2, len(lines)):
                if '"""' in lines[i]:
                    docstring_end = i
                    break
        return (has_docstring, docstring_end)
    
    def _create_docstrings(self, func_indent: int, all_examples: List[str]) -> List[str]:
        # Create the new docstring with all examples
        docstring_lines = []
        docstring_lines.append(func_indent + '    """')
        
        # Add all examples with consistent indentation
        for i, example in enumerate(all_examples):
            # Split example into lines
            example_lines = example.split('\n')
            
            # Add each line with proper indentation
            for j, line in enumerate(example_lines):
                docstring_lines.append(func_indent + '    ' + line)
            
            # Add blank line between examples (not after the last one)
            if i < len(all_examples) - 1:
                docstring_lines.append('')
        
        # Close the docstring
        docstring_lines.append(func_indent + '    """')
        return docstring_lines
        
    def _collect_existing_docstrings(self, func_line_num: int, lines: List[str], has_docstring: bool, docstring_end: int) -> Tuple[List[str], Set[str]]:
        # Collect all examples (existing)
        all_examples = []
        seen_examples = set()
        
        if has_docstring and docstring_end > 0:
            # Extract existing examples from docstring
            existing_examples = self._extract_doctest_examples(lines[func_line_num + 2:docstring_end])
            
            for call, result in existing_examples:
                # Create normalized example
                example = f">>> {call}\n{result}"
                
                if example not in seen_examples:
                    all_examples.append(example)
                    seen_examples.add(example)

        return (all_examples, seen_examples)

    def _add_new_docstrings(self, doctest_examples: List[str], all_examples: List[str], seen_examples: Set[str]) -> Tuple[List[str], Set[str]]:
         # Add new examples
        for example in doctest_examples:
            # Normalize the example format
            parts = example.split('\n', 1)
            if len(parts) == 2:
                call_line, result = parts
                call = call_line[4:].strip()  # Remove >>> prefix and whitespace
                result = result.strip()
                
                # Create normalized example
                normalized_example = f">>> {call}\n{result}"
                
                if normalized_example not in seen_examples:
                    all_examples.append(normalized_example)
                    seen_examples.add(normalized_example)

        return (all_examples, seen_examples)


class DBTestCase:
    def __init__(self, expected_output, inputs, test_function: str, last_run_time, test_method_type: int, test_suite_id: int, function_id: int):
        self._expected_output = expected_output
        self._inputs = inputs
        self._test_function = test_function
        self._last_run_time = last_run_time
        self._test_method_type = test_method_type
        self._test_suite_id = test_suite_id
        self._function_id = function_id

    @property
    def expected_output(self):
        return self._expected_output

    @expected_output.setter
    def expected_output(self, value):
        self._expected_output = value

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, value):
        self._inputs = value

    @property
    def test_function(self) -> str:
        return self._test_function

    @test_function.setter
    def test_function(self, value: str) -> None:
        self._test_function = value

    @property
    def last_run_time(self):
        return self._last_run_time

    @last_run_time.setter
    def last_run_time(self, value) -> None:
        self._last_run_time = value

    @property
    def test_method_type(self) -> int:
        return self._test_method_type

    @test_method_type.setter
    def test_method_type(self, value: int) -> None:
        self._test_method_type = value

    @property
    def test_suite_id(self) -> int:
        return self._test_suite_id

    @test_suite_id.setter
    def test_suite_id(self, value: int) -> None:
        self._test_suite_id = value

    @property
    def function_id(self) -> int:
        return self._function_id

    @function_id.setter
    def function_id(self, value: int) -> None:
        self._function_id = value



from testgen.analyzer.contracts.contract import Contract
from abc import ABC

class NoExceptionContract(Contract, ABC):
    """Ensures function does not raise an exception"""

    def check_preconditions(self, args) -> bool:
        """Returns False if there is a violation"""
        return True

    """Returns False if there is an exception"""
    def check_postconditions(self, args, output, exception) -> bool:
        if exception is not None:
            return True
        else:
            return False


import string
from typing import List

class TestCase:
    def __init__(self, func_name, inputs: tuple, expected: any):
        self.func_name = func_name
        self.inputs = inputs
        self.expected = expected


class Function:
    def __init__(self, name: str, params, start_line: int, end_line: int, num_lines: int, source_file_id: int):
        self._name = name
        self._params = params
        self._start_line = start_line
        self._end_line = end_line
        self._num_lines = num_lines
        self._source_file_id = source_file_id

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def params(self) -> str:
        return self._params

    @params.setter
    def params(self, value: str) -> None:
        self._params = value

    @property
    def start_line(self) -> int:
        return self._start_line

    @start_line.setter
    def start_line(self, value: int) -> None:
        self._start_line = value

    @property
    def end_line(self) -> int:
        return self._end_line

    @end_line.setter
    def end_line(self, value: int) -> None:
        self._end_line = value

    @property
    def num_lines(self) -> int:
        return self._num_lines

    @num_lines.setter
    def num_lines(self, value: int) -> None:
        self._num_lines = value

    @property
    def source_file_id(self) -> int:
        return self._source_file_id

    @source_file_id.setter
    def source_file_id(self, value: int) -> None:
        self._source_file_id = value


import os
from datetime import datetime
from types import ModuleType
from typing import List, Dict, Any
import testgen.util.utils as utils
from testgen.models.db_test_case import DBTestCase
from testgen.models.function import Function
from testgen.models.source_file import SourceFile
from testgen.models.test_case import TestCase

from testgen.db.dao_impl import DaoImpl

class DBService:
    def __init__(self, db_name="testgen.db"):
        self.dao = DaoImpl(db_name)

    def close(self):
        """Close the database connection."""
        self.dao.close()

    def insert_test_suite(self, name: str) -> int:
        return self.dao.insert_test_suite(name)

    def insert_source_file(self, path: str, lines_of_code: int, last_modified) -> int:
        return self.dao.insert_source_file(path, lines_of_code, last_modified)

    def insert_function(self, name: str, params, start_line: int, end_line: int, source_file_id: int) -> int:
        return self.dao.insert_function(name, params, start_line, end_line, source_file_id)

    def insert_test_case(self, test_case: TestCase, test_suite_id: int, function_id: int, test_method_type: int) -> int:
        return self.dao.insert_test_case(test_case, test_suite_id, function_id, test_method_type)

    def insert_test_result(self, test_case_id: int, status: bool, error: str = None) -> int:
        return self.dao.insert_test_result(test_case_id, status, error)

    # TODO: Add support for function_id
    def insert_coverage_data(self, file_name: str, executed_lines: int, missed_lines: int,
                             branch_coverage: float, source_file_id: int, function_id=None) -> int:
        return self.dao.insert_coverage_data(file_name, executed_lines, missed_lines, branch_coverage, source_file_id, function_id=None)

    def get_test_suites(self):
        return self.dao.get_test_suites()

    def get_source_file_id_by_path(self, filepath: str) -> int:
        return self.dao.get_source_file_id_by_path(filepath)

    def get_test_suite_id_by_name(self, name: str) -> int:
        return self.dao.get_test_suite_id_by_name(name)

    def get_functions_by_file(self, filepath: str) -> List[Function]:
        return self.dao.get_functions_by_file(filepath)

    def get_function_id_by_name_file_id_start(self, name: str, source_file_id: int, start_line: int) -> int:
        return self.dao.get_function_by_name_file_id_start(name, source_file_id, start_line)

    def get_test_cases_by_function(self, function_name):
        return self.dao.get_test_cases_by_function(function_name)

    def get_coverage_by_file(self, file_path):
        return self.dao.get_coverage_by_file(file_path)

    def get_test_file_data(self, file_path: str):
        return self.dao.get_test_file_data(file_path)

    def get_test_case_id_by_func_id_input_expected(self, function_id: int, inputs: str, expected: str) -> int:
        return self.dao.get_test_case_id_by_func_id_input_expected(function_id, inputs, expected)

    def save_test_generation_data(self, file_path: str, test_cases: list, test_strategy: int, module: ModuleType, class_name: str | None):
        """Save test generation data to the database."""
        source_file_data = self._get_source_file_data(file_path)
        source_file_id = self.insert_source_file(source_file_data.path, source_file_data.lines_of_code, source_file_data.last_modified)

        test_suite_name = class_name if class_name else module.__name__
        test_suite_id = self.insert_test_suite(test_suite_name)

        function_data = self._get_function_data(file_path)
        for function in function_data:
            self.insert_function(function.name, function.params, function.start_line, function.end_line, source_file_id)

        test_cases_data = self._get_test_cases_data(source_file_id, test_suite_id, function_data, test_cases, test_strategy)
        for test_case in test_cases_data:
            self.insert_test_case(TestCase(test_case.test_function, test_case.inputs, test_case.expected_output), test_case.test_suite_id, test_case.function_id, test_strategy)

    def _get_source_file_data(self, file_path: str) -> SourceFile:
        lines_of_code = sum(1 for _ in open(file_path, 'r')) # Count lines in file
        last_modified_time = os.path.getmtime(file_path)
        return SourceFile(file_path, lines_of_code, last_modified_time)

    def _get_function_data(self, file_path: str) -> List[Function]:
        return utils.get_list_of_functions(file_path)

    def _get_test_cases_data(self, source_file_id: int, test_suite_id: int, function_data: List[Function], test_cases: List[TestCase], test_strategy: int) -> List[DBTestCase]:
        db_test_cases = []
        for test_case in test_cases:
            function_id = self.match_test_case_to_function_for_id(source_file_id, test_case, function_data)
            db_test_case = DBTestCase(test_case.expected, test_case.inputs, test_case.func_name, datetime.now(), test_strategy, test_suite_id, function_id)
            db_test_cases.append(db_test_case)
        return db_test_cases

    def match_test_case_to_function_for_id(self, source_file_id: int, test_case: TestCase, functions: List[Function]) -> int:
        func_name = test_case.func_name
        if "." in func_name:
            func_name = func_name.split(".")[-1]

        candidate_functions = [f for f in functions if f.name.endswith(func_name)]

        if not candidate_functions:
            return -1

        if len(candidate_functions) == 1:
            return self.get_function_id_by_name_file_id_start(candidate_functions[0].name, source_file_id, candidate_functions[0].start_line)

        # Match by parameter count
        input_param_count = len(test_case.inputs) if isinstance(test_case.inputs, dict) else 1

        for function in candidate_functions:
            # Parse params from string representation to dict if needed
            params = function.params
            if isinstance(params, str):
                try:
                    params = eval(params)  # Convert string representation to dict
                except (SyntaxError, NameError):
                    continue

            param_count = len(params) if isinstance(params, dict) else 0

            if param_count == input_param_count:
                return self.get_function_id_by_name_file_id_start(function.name, source_file_id, function.start_line)

        return -1

    def _get_test_results(self, filepath: str, execution_results, test_format: int):
        source_file_id = self.get_source_file_id_by_path(filepath)
        functions = self.get_functions_by_file(filepath)
        for result in execution_results:
            name = result.name
            test_case = utils.parse_test_case_from_result_name(name, test_format)
            function_id = self.match_test_case_to_function_for_id(source_file_id, test_case, functions)
            test_case_id = self.get_test_case_id_by_func_id_input_expected(function_id, str(test_case.inputs), test_case.expected)
            self.insert_test_result(test_case_id, result.status, result.error)


class TestResult:
    def __init__(self, test_case_id: int, status: bool, error: str, execution_time):
        self._test_case_id = test_case_id
        self._status = status
        self._error = error
        self._execution_time = execution_time

    @property
    def test_case_id(self) -> int:
        return self._test_case_id

    @test_case_id.setter
    def test_case_id(self, value: int) -> None:
        self._test_case_id = value

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: bool) -> None:
        self._status = value

    @property
    def error(self) -> str:
        return self._error

    @error.setter
    def error(self, value: str) -> None:
        self._error = value

    @property
    def execution_time(self):
        return self._execution_time

    @execution_time.setter
    def execution_time(self, value) -> None:
        self._execution_time = value


import os
import subprocess
import time

from testgen.models.coverage_data import CoverageData
from testgen.service.logging_service import get_logger
from testgen.service.db_service import DBService

class CoverageService:
    UNITTEST_FORMAT = 1
    PYTEST_FORMAT = 2
    DOCTEST_FORMAT = 3

    def __init__(self):
        self.logger = get_logger()

    def run_coverage(self, test_file: str, source_file: str) -> CoverageData:
        #self.wait_for_file(test_file)
        self.logger.info(f"Running coverage on test file: {test_file}")
        self.logger.info(f"Source file to measure: {source_file}")

        try:
            cov_data = self.collect_coverage(test_file, source_file)
            return cov_data
        except Exception as e:
            self.logger.error(f"Error running coverage: {str(e)}")
            raise RuntimeError(f"Error running coverage subprocess: {e}")

    def collect_coverage(self, test_file: str, source_file: str) -> CoverageData:
        try:
            subprocess.run(["python", "-m", "coverage", "run", "--source=.", test_file], check=True)
            result = subprocess.run(
                ["python", "-m", "coverage", "report", source_file],
                check=True,
                capture_output=True,
                text=True
            )
            coverage_output = result.stdout
            subprocess.run(["python", "-m", "coverage", "json"], check=True)
            return self.parse_coverage_data(coverage_output, source_file)
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Error collecting coverage: {str(e)}")
            return CoverageData(
                coverage_type="file",
                executed_lines=0,
                missed_lines=0,
                branch_coverage=0.0,
                source_file_id=-1,
                function_id=None
            )

    def parse_coverage_data(self, coverage_output: str, file_path: str) -> CoverageData:
        lines = coverage_output.strip().split('\n')
        executed_lines = missed_lines = total_lines = 0
        branch_coverage = 0.0

        if lines:
            for line in lines:
                if file_path in line:
                    parts = line.split()
                    if len(parts) >= 4:
                        try:
                            total_lines = int(parts[-3])
                            missed_lines = int(parts[-2])
                            executed_lines = total_lines - missed_lines
                            coverage_str = parts[-1].strip('%')
                            branch_coverage = float(coverage_str) / 100
                            break
                        except (ValueError, IndexError) as e:
                            self.logger.error(f"Error parsing coverage data: {e}")

        # You may want to look up the source_file_id here if needed
        source_file_id = -1  # Set this appropriately in your context

        return CoverageData(
            coverage_type="file",
            executed_lines=executed_lines,
            missed_lines=missed_lines,
            branch_coverage=branch_coverage,
            source_file_id=source_file_id,
            function_id=None
        )

    # Save coverage data to database currently not working
    """"
    def save_coverage_data(self, db_service: DBService, coverage_data: CoverageData, file_path: str) -> None:
        if db_service is None:
            self.logger.debug("Skipping database operations - no DB service provided")
            return

        try:
            source_file_id = db_service.get_source_file_id_by_path(file_path)
            if source_file_id == -1:
                self.logger.error(f"Source file not found in database: {file_path}")
                return

            db_service.insert_coverage_data(
                file_name=file_path,
                executed_lines=coverage_data.executed_lines,
                missed_lines=coverage_data.missed_lines,
                branch_coverage=coverage_data.branch_coverage,
                source_file_id=source_file_id
            )
        except Exception as e:
            self.logger.error(f"Error saving coverage data to database: {e}")
    """

    @staticmethod
    def wait_for_file(file_path, retries=5, delay=1):
        """Wait for the generated file to appear."""
        while retries > 0 and not os.path.exists(file_path):
            time.sleep(delay)
            retries -= 1
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File '{file_path}' not found after waiting.")


import ast
from abc import ABC, abstractmethod
from typing import List, Dict

from testgen.models.test_case import TestCase
from testgen.models.analysis_context import AnalysisContext
from testgen.models.function_metadata import FunctionMetadata
from testgen.service.logging_service import get_logger

class TestCaseAnalyzerStrategy(ABC):
    def __init__(self, analysis_context: AnalysisContext = None):
        self._analysis_context = analysis_context
        self.logger = get_logger()

    @abstractmethod
    def collect_test_cases(self, function_metadata: FunctionMetadata) -> List[TestCase]:
        pass

    def get_function_metadata(self, func_name: str) -> FunctionMetadata | None:
        if self._analysis_context and self._analysis_context.function_data:
            for func_metadata in self._analysis_context.function_data:
                if func_metadata.function_name == func_name:
                    return func_metadata
        return None

    def get_function_metadata_by_node(self, func_node: ast.FunctionDef) -> FunctionMetadata:
        return self.get_function_metadata(func_node.name)

    def get_param_types(self, func_node: ast.FunctionDef) -> Dict[str, str] | None:
        func_metadata = self.get_function_metadata_by_node(func_node)
        if func_metadata:
            return func_metadata.params

        return None

    @property
    def analysis_context(self) -> AnalysisContext:
        return self._analysis_context
        
    @analysis_context.setter
    def analysis_context(self, context: AnalysisContext):
        self._analysis_context = context
        

    

    




from testgen.service.service import Service
from testgen.controller.cli_controller import CLIController
from testgen.generator.unit_test_generator import UnitTestGenerator

def main():
    service = Service()
    controller = CLIController(service)
    controller.run()

if __name__ == '__main__':
    main()


import argparse
import os
import sys

import docker
from docker import DockerClient
from docker import errors
from testgen.service.logging_service import LoggingService, get_logger
from testgen.util.file_utils import adjust_file_path_for_docker, get_project_root_in_docker
from testgen.controller.docker_controller import DockerController
from testgen.service.service import Service
from testgen.service.db_service import DBService

AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3
REINFORCE_STRAT = 4

UNITTEST_FORMAT = 1
PYTEST_FORMAT = 2
DOCTEST_FORMAT = 3

class CLIController:
    #TODO: Possibly create a view 'interface' and use dependency injection to extend other views
    def __init__(self, service: Service):
        self.service = service
        self.logger = None

    def run(self):

        parser = self.add_arguments()

        args = parser.parse_args()

        LoggingService.get_instance().initialize(
            debug_mode=args.debug if args.debug else False,
            log_file=args.log_file if hasattr(args, 'log_file') else None,
            console_output=True
        )

        self.logger = get_logger()

        if args.functions:
            self.service.get_all_functions(args.file_path)
            return

        if args.query:
            print(f"Querying database for file: {args.file_path}")
            self.service.query_test_file_data(args.file_path)
            return
        
        if args.coverage:
            self.service.run_coverage(args.file_path)
            return
            
        running_in_docker = os.environ.get("RUNNING_IN_DOCKER") is not None
        if running_in_docker:
            args.file_path = adjust_file_path_for_docker(args.file_path)
            self.execute_generation(args, True)
        elif args.safe and not running_in_docker:
            client = self.docker_available()
            # Skip Docker-dependent operations if client is None
            if client is None and args.safe:
                self.logger.debug("Running with --safe flag requires Docker. Continuing without safe mode.")
                args.safe = False
                self.execute_generation(args)
            else:
                docker_controller = DockerController()
                project_root = get_project_root_in_docker(args.file_path)
                successful: bool = docker_controller.run_in_docker(project_root, client, args)
                if not successful:
                    if hasattr(args, 'db') and args.db:
                        self.service.db_service = DBService(args.db)
                        self.logger.debug(f"Using database: {args.db}")
                    self.execute_generation(args)
        else:
            if hasattr(args, 'db') and args.db:
                self.service.db_service = DBService(args.db)
                self.logger.debug(f"Using database: {args.db}")
            self.logger.debug("Running in local mode...")
            self.execute_generation(args)

    def execute_generation(self, args: argparse.Namespace, running_in_docker: bool = False):
        try:
            self.set_service_args(args)

            if running_in_docker:
                self.logger.debug("Running in Docker mode...")
                self.service.generate_test_cases()

            else:
                test_file = self.service.generate_tests(args.output)
                self.logger.debug(f"Unit tests saved to: {test_file}")
                print("Executing tests...")
                self.service.run_tests(test_file)
                print("Running coverage...")
                self.service.run_coverage(test_file)
                self.logger.debug("Tests and coverage data saved to database.")

                if args.visualize:
                    self.service.visualize_test_coverage()

        except Exception as e:
            self.logger.error(f"An error occurred: {e}")
            if hasattr(self.service, 'db_service'):
                self.service.db_service.close()

    def set_service_args(self, args: argparse.Namespace):
        self.service.set_file_path(args.file_path)
        self.set_test_format(args)
        self.set_test_strategy(args)

    def add_arguments(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(description="A CLI tool for generating unit tests.")
        parser.add_argument("file_path", type=str, help="Path to the Python file.")
        parser.add_argument("--output", "-o", type=str, help="Path to output directory.")
        parser.add_argument("-q", "--query", action="store_true", help="Query the database for test cases, coverage data, and test results for a specific file")
        parser.add_argument(
            "--generate-only", "-g",
            action="store_true",
            help="Generate branched code but skip running unit tests and coverage."
        )
        parser.add_argument(
            "--test-mode",
            choices=["ast", "random", "fuzz", "reinforce"],
            default="ast",
            help="Set the test generation analysis technique"
        )
        parser.add_argument(
            "--reinforce-mode",
            choices=["train", "collect"],
            default="train",
            help="Set mode for reinforcement learning"
        )
        parser.add_argument(
            "--test-format",
            choices=["unittest", "pytest", "doctest"],
            default="unittest",
            help="Set the test generation format"
        )
        parser.add_argument(
            "--safe",
            action="store_true",
            help="Run test generation from within a docker container."
        )
        parser.add_argument(
            "--db",
            type=str,
            default="testgen.db",
            help="Path to SQLite database file (default: testgen.db)"
        )
        parser.add_argument(
            "-viz", "--visualize",
            action="store_true",
            help = "Visualize the tests with graphviz"
        )
        parser.add_argument(
            "--debug",
            action="store_true",
            help="Enable debug logging"
        )
        parser.add_argument(
            "--log-file",
            type=str,
            help="Path to log file (if not specified, logs will only go to console)"
        )
        parser.add_argument(
            "-c", "--coverage",
            action="store_true",
            help="Run coverage analysis on the generated tests"
        )
        parser.add_argument(
            "-f", "--functions",
            action="store_true",
            help="List all functions in file"
        )
        return parser

    def set_test_format(self, args: argparse.Namespace):
        if args.test_format == "pytest":
            self.service.set_test_generator_format(PYTEST_FORMAT)
        elif args.test_format == "doctest":
            self.service.set_test_generator_format(DOCTEST_FORMAT)
        else:
            self.service.set_test_generator_format(UNITTEST_FORMAT)

    def set_test_strategy(self, args: argparse.Namespace):
        if args.test_mode == "random":
            print("Using Random Feedback-Directed Test Generation Strategy.")
            self.service.set_test_analysis_strategy(RANDOM_STRAT)
        elif args.test_mode == "fuzz":
            print("Using Fuzz Test Generation Strategy...")
            self.service.set_test_analysis_strategy(FUZZ_STRAT)
        elif args.test_mode == "reinforce":
            print("Using Reinforcement Learning Test Generation Strategy...")
            if args.reinforce_mode == "train":
                print("Training mode enabled - will update Q-table")
            else:
                print("Training mode disabled - will use existing Q-table")
            self.service.set_test_analysis_strategy(REINFORCE_STRAT)
            self.service.set_reinforcement_mode(args.reinforce_mode)
        else:
            print("Generating function code using AST analysis...")
            generated_file_path = self.service.generate_function_code()
            print(f"Generated code saved to: {generated_file_path}")
            if not args.generate_only:
                print("Using Simple AST Traversal Test Generation Strategy...")
                self.service.set_test_analysis_strategy(AST_STRAT)

    def docker_available(self) -> DockerClient | None:
        try:
            client = docker.from_env()
            client.ping()
            print("Docker daemon is running and connected.")
            return client
        except docker.errors.DockerException as err:
            print(f"Docker is not available: {err}")
            print(f"Make sure the Docker daemon is running, and try again.")
            choice = input("Continue without Docker (y/n)?")
            if choice.lower() == 'y':
                return None
            else:
                sys.exit(1)



import inspect
import json
import os
import re
import time
import subprocess

from types import ModuleType
from typing import Any, Dict, List

import testgen.util.file_utils as file_utils
import testgen.util.utils
from testgen.models.coverage_data import CoverageData
from testgen.models.function import Function

from testgen.models.test_case import TestCase
from testgen.service.analysis_service import AnalysisService
from testgen.service.generator_service import GeneratorService
from testgen.service.coverage_service import CoverageService
from testgen.service.db_service import DBService
from testgen.models.analysis_context import AnalysisContext
from testgen.service.logging_service import get_logger
from testgen.service.test_executor_service import TestExecutorService

# Constants for test strategies
AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3
REINFORCE_STRAT = 4

# Constants for test formats
UNITTEST_FORMAT = 1
PYTEST_FORMAT = 2
DOCTEST_FORMAT = 3

class Service:
    def __init__(self):
        self.test_strategy: int = 0
        self.test_format: int = 0
        self.file_path = None
        self.generated_file_path = None
        self.class_name = None
        self.test_cases = []
        self.logger = get_logger()
        self.reinforcement_mode = "train"
        
        # Initialize specialized services
        self.analysis_service = AnalysisService()
        self.generator_service = GeneratorService(None, None, None)
        self.coverage_service = CoverageService()
        self.test_executor_service = TestExecutorService()
        self.coverage_service = CoverageService()
        # Only initialize DB service if not running in Docker
        if os.environ.get("RUNNING_IN_DOCKER") is None:
            self.db_service = DBService()
        else:
            # Create a dummy DB service that doesn't do anything
            self.db_service = None

    def generate_test_cases(self) -> List[TestCase] | None:
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)

        if self.test_strategy == AST_STRAT:
            self.generated_file_path = self.generate_function_code()
            Service.wait_for_file(self.generated_file_path)
            self.analysis_service.set_file_path(self.generated_file_path)
            module = file_utils.load_module(self.generated_file_path)
            class_name = self.analysis_service.get_class_name(module)
            self.analysis_service.set_test_strategy(self.test_strategy, module.__name__, self.class_name)

        else:
            self.analysis_service.set_test_strategy(self.test_strategy, module.__name__, self.class_name)

        test_cases: List[TestCase] = []
        if self.test_strategy == REINFORCE_STRAT:
            test_cases = self.analysis_service.do_reinforcement_learning(self.file_path, class_name, self.reinforcement_mode)
        else:
            test_cases = self.analysis_service.generate_test_cases()

        if os.environ.get("RUNNING_IN_DOCKER") is not None:
            self.logger.debug(f"Serializing test cases {test_cases}")
            self.serialize_test_cases(test_cases)
            return None  # Exit early in analysis-only mode
        
        return test_cases

    def generate_tests(self, output_path=None):
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)

        test_cases = self.generate_test_cases()
        
        # Only process if we have test cases
        if test_cases is None:
            return None
            
        self.test_cases = test_cases

        # Only save to DB if not running in Docker
        if os.environ.get("RUNNING_IN_DOCKER") is None:
            file_path_to_use = self.generated_file_path if self.test_strategy == AST_STRAT else self.file_path
            # Don't save test to db foreign key constraint violation error
            """self.db_service.save_test_generation_data(
                file_path_to_use,
                test_cases,
                self.test_strategy,
                module,
                class_name
            )"""

        test_file = self.generate_test_file(test_cases, output_path, module, class_name)

        # Ensure the test file is ready
        Service.wait_for_file(test_file)
        return test_file

    def generate_test_file(self, test_cases: List[TestCase], output_path: str | None = None,
                           module: ModuleType | None = None, class_name: str | None = None) -> str:
        if module is None:
            module = file_utils.load_module(self.file_path)
            class_name = self.analysis_service.get_class_name(module)

        if self.test_strategy == AST_STRAT:
            self.generator_service = GeneratorService(self.generated_file_path, output_path, self.test_format)
            self.generator_service.set_test_format(self.test_format)
            module = file_utils.load_module(self.generated_file_path)
        else:
            # Create the correct instance of the generator service
            self.generator_service = GeneratorService(self.file_path, output_path, self.test_format)
            self.generator_service.set_test_format(self.test_format)

        test_file = self.generator_service.generate_test_file(
            module,
            class_name,
            test_cases,
            output_path
        )

        print(f"Generated test file: {test_file}")

        # Ensure the test file is ready
        Service.wait_for_file(test_file)
        return test_file

    def generate_function_code(self):
        """Generate function code for a given class or module."""
        module = file_utils.load_module(self.file_path)
        class_name = self.analysis_service.get_class_name(module)
        functions = self.inspect_class(class_name)
        return self.generator_service.generate_function_code(self.file_path, class_name, functions)

    def run_tests(self, test_file: str):
        # Run execute tests, would collect results but currently not saving them to db
        _ = self.test_executor_service.execute_tests(test_file, self.test_format)
        """
        if results is None:
            raise RuntimeError("No test results returned from the test executor service.")
        else:
            if self.db_service:
                self.test_executor_service.save_test_results(self.db_service, results, self.file_path, self.test_format)
        """

    def run_coverage(self, test_file: str):
        file_path_to_use = self.generated_file_path if self.test_strategy == AST_STRAT else self.file_path

        self.logger.debug(f"Running coverage on: {test_file}")
        coverage_data = self.coverage_service.run_coverage(test_file, file_path_to_use)

        if os.environ.get("RUNNING_IN_DOCKER") is not None or self.db_service is None:
            self.logger.debug("Skipping database operations - running in Docker or no DB service")
            return

        # Save test results and coverage data
        """self.coverage_service.save_coverage_data(
            db_service=self.db_service,
            coverage_data=coverage_data,
            file_path=file_path_to_use
        )"""

        self._print_coverage_summary(file_path_to_use, coverage_data)

    @staticmethod
    def _print_coverage_summary(file_path: str, coverage_data: CoverageData):
        print("\nCoverage Summary:")
        print(f"File: {file_path}")
        total_lines = coverage_data.missed_lines + coverage_data.executed_lines
        print(f"Total lines: {total_lines}")
        print(f"Executed lines: {coverage_data.executed_lines}")
        print(f"Missed lines: {coverage_data.missed_lines}")
        
        if total_lines > 0:
            percentage = (coverage_data.executed_lines / total_lines) * 100
            print(f"Coverage: {percentage:.2f}%")
        else:
            print("Coverage: N/A (no lines to cover)")

    def serialize_test_cases(self, test_cases):
        """Serialize input arguments ot JSON-compatible format"""
        print("##TEST_CASES_BEGIN##")
        serialized = []
        for tc in test_cases:
            case_data = {
                "func_name": tc.func_name,
                "inputs": self.serialize_value(tc.inputs),
                "expected": self.serialize_value(tc.expected)
            }
            serialized.append(case_data)
        print(json.dumps(serialized))
        print("##TEST_CASES_END##")

    def serialize_value(self, value):
        """Serialize a single value to a JSON-compatible format"""
        if value is None or isinstance(value, (bool, int, float, str)):
            return value
        elif isinstance(value, (list, tuple)):
            return [self.serialize_value(v) for v in value]
        else:
            return str(value)

    @staticmethod
    def parse_test_cases_from_logs(logs_output):
        """Extract and parse test cases from container logs"""
        pattern = r"##TEST_CASES_BEGIN##\n(.*?)\n##TEST_CASES_END##"
        match = re.search(pattern, logs_output, re.DOTALL)

        if not match:
            raise ValueError("Could not find test cases in the container logs")

        test_cases_json = match.group(1).strip()
        test_cases_data = json.loads(test_cases_json)

        test_cases: List[TestCase] = []
        for tc_data in test_cases_data:
            test_case = TestCase(
                func_name=tc_data["func_name"],
                inputs=tc_data["inputs"],
                expected=tc_data["expected"]
            )
            test_cases.append(test_case)

        return test_cases

    def set_file_path(self, path: str):
        """Set the file path for analysis and validate it."""
        print(f"Setting file path: {path}")
        if os.path.isfile(path) and path.endswith(".py"):
            self.file_path = path
            self.analysis_service.set_file_path(path)
        else:
            raise ValueError("Invalid file path! Please provide a valid Python file path.")

    def set_class_name(self, class_name: str):
        """Set the class name to analyze."""
        self.class_name = class_name

    def set_test_generator_format(self, test_format: int):
        """Set the test generator format."""
        self.test_format = test_format
        self.generator_service.set_test_format(test_format)

    def set_test_analysis_strategy(self, strategy: int):
        """Set the test analysis strategy."""
        self.test_strategy = strategy
        module = file_utils.load_module(self.file_path)
        self.analysis_service.set_test_strategy(strategy, module.__name__, self.class_name)
    
    def get_analysis_context(self, filepath: str) -> AnalysisContext:
        """Create an analysis context for the given file."""
        return self.analysis_service.create_analysis_context(filepath)

    @staticmethod
    def wait_for_file(file_path, retries=5, delay=1):
        """Wait for the generated file to appear."""
        while retries > 0 and not os.path.exists(file_path):
            time.sleep(delay)
            retries -= 1
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File '{file_path}' not found after waiting.")

    def find_package_root(self):
        """Find the package root directory."""
        current_dir = os.path.abspath(os.path.dirname(self.file_path))
        last_valid = None

        while current_dir:
            if "__init__.py" in os.listdir(current_dir):
                last_valid = current_dir
            else:
                break

            parent_dir = os.path.dirname(current_dir)
            if parent_dir == current_dir:
                break
            current_dir = parent_dir

        return last_valid

    def inspect_class(self, class_name):
        """Inspect a class or module and return its functions."""
        module = file_utils.load_module(self.file_path)

        # Handle module-level functions when class_name is None
        if not class_name:
            # Get module-level functions
            functions = inspect.getmembers(module, inspect.isfunction)
            return functions

        # Handle class functions
        cls = getattr(module, class_name, None)
        if cls is None:
            raise ValueError(f"Class '{class_name}' not found in module '{module.__name__}'.")

        functions = inspect.getmembers(cls, inspect.isfunction)
        return functions

    def resolve_module_path(self, module_name):
        """Resolve a module name to its file path by checking multiple locations."""
        direct_path = f"/controller/{module_name}.py"
        if os.path.exists(direct_path):
            self.logger.debug(f"Found module at {direct_path}")
            return direct_path

        testgen_path = f"/controller/testgen/{module_name}.py"
        if os.path.exists(testgen_path):
            self.logger.debug(f"Found module at {testgen_path}")
            return testgen_path

        if '.' in module_name:
            parts = module_name.split('.')
            potential_path = os.path.join('/controller', *parts) + '.py'
            if os.path.exists(potential_path):
                self.logger.debug(f"Found module at {potential_path}")
                return potential_path

        self.logger.debug(f"Could not find module: {module_name}")
        return None

    def visualize_test_coverage(self):
        from testgen.service.cfg_service import CFGService
        cfg_service = CFGService()
        cfg_service.initialize_visualizer(self)

        return cfg_service.visualize_test_coverage(
            file_path=self.file_path,
            test_cases=self.test_cases,
        )

    def set_reinforcement_mode(self, mode: str):
        self.reinforcement_mode = mode
        if hasattr(self ,'analysis_service'):
            self.analysis_service.set_reinforcement_mode(mode)

    def query_test_file_data(self, test_file_name: str):
        if self.db_service is None:
            raise RuntimeError("Database service is not initialized.")

        results = self.db_service.get_test_file_data(test_file_name)
        if not results:
            self.logger.debug(f"No data found for file: {test_file_name}")
            return

        from tabulate import tabulate
        rows = [dict(row) for row in results]

        print(f"Results for file: {test_file_name}")
        print(tabulate(rows, headers="keys", tablefmt="grid"))

    def get_all_functions(self, file_path: str):
        functions = testgen.util.utils.get_list_of_functions(file_path)
        for func in functions:
            print(f"Function: {func.name}")
            for attr, value in vars(func).items():
                if attr == "_source_file_id":
                    continue
                else:
                    print(f"  {attr}: {value}")
        return

    def select_all_from_db(self) -> None:
        rows = self.db_service.get_test_suites()
        for row in rows:
            print(repr(dict(row)))



import ast
from typing import List

import testgen.util.file_utils
from testgen.models.test_case import TestCase
from testgen.analyzer.test_case_analyzer import TestCaseAnalyzerStrategy
from abc import ABC
from testgen.models.function_metadata import FunctionMetadata

class ASTAnalyzer(TestCaseAnalyzerStrategy, ABC):
    def __init__(self, analysis_context=None):
        super().__init__(analysis_context)

    def collect_test_cases(self, function_metadata: FunctionMetadata) -> List[TestCase]:
        """Collect test cases by analyzing AST conditions and return statements"""

        if function_metadata and function_metadata.params:
            param_names = list(function_metadata.params.keys())
        else:
            param_names = [arg.arg for arg in function_metadata.func_def.args.args
                           if arg.arg != 'self']

        test_cases: List[TestCase] = self.get_conditions_recursively(
            function_metadata,
            function_metadata.function_name,
            function_metadata.func_def.body,
            param_names)
        return test_cases

    def get_conditions_recursively(self, function_metadata: FunctionMetadata, func_name: str, func_node_body: list, param_names, test_cases=None, conditions=None) -> List[TestCase]:
        if conditions is None:
            conditions = []

        if test_cases is None:
            test_cases = []

        for node in func_node_body:
            if isinstance(node, ast.If):
                condition_str = self.parse_condition(node.test)
                self.logger.debug(f"Condition found in function: {condition_str}")
                self.get_conditions_recursively(function_metadata, func_name, node.body, param_names, test_cases, conditions + [condition_str])
                if node.orelse:
                    self.get_conditions_recursively(function_metadata, func_name, node.orelse, param_names, test_cases, conditions)

            elif isinstance(node, ast.Return):
                inputs = self.generate_inputs_from_conditions(conditions, param_names)
                
                input_exists = False
                for existing_test in test_cases:
                    if existing_test.func_name == func_name and existing_test.inputs == inputs:
                        input_exists = True
                        break
                    
                if not input_exists:
                    try:
                        func = function_metadata.func
                        output = func(*inputs)
                        
                    except Exception as e:
                        print(f"Error executing function: {e}")
                        output = self._try_to_determine_output(node)
                    
                    test_case = TestCase(func_name, inputs, output)
                    test_cases.append(test_case)

        return test_cases

    def _try_to_determine_output(self, return_node):
        """Try different strategies to determine the output value"""
        try:
            # Try direct evaluation
            return ast.literal_eval(return_node.value)
        except (ValueError, SyntaxError):
            # Check for constants
            if isinstance(return_node.value, ast.Constant):
                return return_node.value.value
            # Check for name values
            elif isinstance(return_node.value, ast.Name):
                if return_node.value.id == 'True':
                    return True
                elif return_node.value.id == 'False':
                    return False
            # Default fallback
            return None

    @staticmethod
    def parse_condition(condition_node):
        """Parse an AST condition node into a string representation."""
        try:
            if isinstance(condition_node, ast.Compare):
                left = condition_node.left.id if isinstance(condition_node.left, ast.Name) else str(condition_node.left)
                op = type(condition_node.ops[0]).__name__
                right = condition_node.comparators[0].value if isinstance(condition_node.comparators[0], ast.Constant) else str(condition_node.comparators[0])
                return f"{left} {op} {right}"
            elif isinstance(condition_node, ast.BoolOp):
                # Use a simple representation for boolean operations
                return f"BoolOp({type(condition_node.op).__name__})"
            elif isinstance(condition_node, ast.UnaryOp):
                # Use a simple representation for unary operations
                return f"UnaryOp({type(condition_node.op).__name__})"
            elif isinstance(condition_node, ast.Name):
                return condition_node.id
            elif isinstance(condition_node, ast.Constant):
                return str(condition_node.value)
            
            return ast.dump(condition_node)
        except Exception as e:
            print(f"Error parsing condition: {e}")
            return "UnknownCondition"

    @staticmethod
    def generate_inputs_from_conditions(conditions, param_names):
        inputs = []

        # Create parameter values with more variety based on condition path
        param_values = {}

        # If no meaningful conditions were found, create varied inputs
        if not any(param in cond for cond in conditions for param in param_names):
            for i, param in enumerate(param_names):
                # Alternate True/False to create diverse test cases
                param_values[param] = (i % 2 == 0)
        else:
            # Start with all False
            param_values = {param: False for param in param_names}

            # Process each condition to extract parameter values
            for cond in conditions:
                for param in param_names:
                    if param in cond:
                        if "Eq True" in cond or "== True" in cond:
                            param_values[param] = True
                        elif "Eq False" in cond or "== False" in cond:
                            param_values[param] = False
                        elif f"Not({param})" in cond:
                            param_values[param] = False

        # Build the input tuple
        for param in param_names:
            inputs.append(param_values[param])

        return tuple(inputs)


import logging
import os
import sys
from enum import Enum
from typing import Optional


class LogLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class LoggingService:
    """Centralized logging service for testgen framework"""
    _instance = None
    _initialized = False
    
    @classmethod
    def get_instance(cls):
        """Get singleton instance of LoggingService"""
        if cls._instance is None:
            cls._instance = LoggingService()
        return cls._instance
    
    def __init__(self):
        self.logger = logging.getLogger('testgen')
        self.formatter = logging.Formatter(
            '%(asctime)s [%(levelname)s] %(message)s', 
            datefmt='%Y-%m-%d %H:%M:%S'
        )
        self.file_handler = None
        self.console_handler = None
        
    def initialize(self, 
                  debug_mode: bool = False, 
                  log_file: Optional[str] = None,
                  console_output: bool = True):
        """Initialize the logging service"""
        if LoggingService._initialized:
            return
        
        self.debug_mode = debug_mode
            
        # Set the base logging level
        level = LogLevel.DEBUG.value if debug_mode else LogLevel.INFO.value
        self.logger.setLevel(level)
        
        # Add console handler if requested
        if console_output:
            self.console_handler = logging.StreamHandler(sys.stdout)
            self.console_handler.setFormatter(self.formatter)
            self.console_handler.setLevel(level)
            self.logger.addHandler(self.console_handler)
        
        # Add file handler if path provided
        if log_file:
            # Ensure directory exists
            os.makedirs(os.path.dirname(log_file), exist_ok=True)
            self.file_handler = logging.FileHandler(log_file)
            self.file_handler.setFormatter(self.formatter)
            self.file_handler.setLevel(level)
            self.logger.addHandler(self.file_handler)
        
        # Mark as initialized
        LoggingService._initialized = True
        self.info(f"Logging initialized - Debug mode: {debug_mode}")
    
    def debug(self, message: str):
        """Log debug message"""
        if self.debug_mode:
            self.logger.debug(message)
    
    def info(self, message: str):
        """Log info message"""
        self.logger.info(message)
    
    def warning(self, message: str):
        """Log warning message"""
        self.logger.warning(message)
    
    def error(self, message: str):
        """Log error message"""
        self.logger.error(message)
    
    def critical(self, message: str):
        """Log critical message"""
        self.logger.critical(message)


# Global accessor function for easy import and use
def get_logger():
    """Get the global logger instance"""
    logger = LoggingService.get_instance()
    
    # If logger hasn't been initialized yet, set up a basic configuration
    """if not LoggingService._initialized:
        logger.initialize(debug_mode=False, console_output=True)"""
        
    return logger



import ast
from abc import ABC
import random
from typing import List

import testgen.util.randomizer
from testgen.models.function_metadata import FunctionMetadata
from testgen.models.test_case import TestCase
from testgen.analyzer.test_case_analyzer import TestCaseAnalyzerStrategy
from testgen.reinforcement.environment import ReinforcementEnvironment


# Goal: Learn a policy to generate a set of test cases with optimal code coverage, and minimum number of test cases
# Environment: FUT (Function Under Test)
# Agent: System
# Actions: Create new test case, combine test cases, delete test cases
# Rewards:

from typing import List, Optional
from testgen.models.test_case import TestCase
from testgen.models.analysis_context import AnalysisContext
from testgen.analyzer.test_case_analyzer import TestCaseAnalyzerStrategy
from testgen.reinforcement.agent import ReinforcementAgent
from testgen.reinforcement.environment import ReinforcementEnvironment
from testgen.reinforcement.statement_coverage_state import StatementCoverageState


class ReinforcementAnalyzer(TestCaseAnalyzerStrategy):
    def __init__(self, analysis_context: AnalysisContext, mode: str = "train"):
        super().__init__(analysis_context)
        self.analysis_context = analysis_context
        self.mode = mode

    def collect_test_cases(self, function_metadata: FunctionMetadata):
        # Implement or delegate as needed
        return self.analyze(function_metadata)

    def analyze(self, function_metadata: FunctionMetadata) -> List[TestCase]:
        from testgen.service.analysis_service import AnalysisService

        q_table = AnalysisService._load_q_table()
        function_test_cases: List[TestCase] = []


        environment = ReinforcementEnvironment(
            self.analysis_context.filepath,
            function_metadata,
            function_test_cases,
            state=StatementCoverageState(None)
        )
        environment.state = StatementCoverageState(environment)
        agent = ReinforcementAgent(
            self.analysis_context.filepath,
            environment,
            function_test_cases,
            q_table
        )
        episodes = 10 if self.mode == "train" else 1
        for _ in range(episodes):
            if self.mode == "train":
                new_test_cases = agent.do_q_learning()
            else:
                new_test_cases = agent.collect_test_cases()
            function_test_cases.extend(new_test_cases)

        seen = set()
        unique_test_cases = []
        for case in function_test_cases:
            case_inputs = tuple(case.inputs) if isinstance(case.inputs, list) else case.inputs
            case_key = (case.func_name, case_inputs)
            if case_key not in seen:
                seen.add(case_key)
                unique_test_cases.append(case)

        AnalysisService._save_q_table(q_table)
        return unique_test_cases



import ast
import os
import string
import sys
import random
from typing import List

from atheris import FuzzedDataProvider

import testgen.util.file_utils
from testgen.models.function import Function
from testgen.models.test_case import TestCase
from testgen.util.file_utils import load_and_parse_file_for_tree

def extract_parameter_types(func_node):
    """Extract parameter types from a function node."""
    param_types = {}
    for arg in func_node.args.args:
        param_name = arg.arg
        if arg.annotation:
            param_type = ast.unparse(arg.annotation)
            param_types[param_name] = param_type
        else:
            if param_name != 'self':
                param_types[param_name] = None
    return param_types

def generate_random_inputs(param_types):
    """Generate inputs for fuzzing based on parameter types."""
    inputs = {}
    for param, param_type in param_types.items():
        if param_type == "int":
            # Try using sys.maxsize for minsize and maxsize
            minsize = -sys.maxsize - 1
            maxsize = sys.maxsize
            random_integer = random.randint(minsize, maxsize)
            inputs[param] = random_integer
        elif param_type == "bool":
            random_choice = random.choice([True, False])
            inputs[param] = random_choice
        elif param_type == "float":
            random_float = random.random()
            inputs[param] = random_float
        # TODO: Random String and Random bytes; Random objects?
        elif param_type == "str":
            inputs[param] = "abc"
        #elif param_type == "bytes":
        #    inputs[param] = fdp.ConsumeBytes(10)
        else:
            inputs[param] = None
    return inputs

def get_functions(tree) -> List[ast.FunctionDef]:
    functions = []
    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            functions.append(node)
        if isinstance(node, ast.ClassDef):
            for class_node in node.body:
                if isinstance(class_node, ast.FunctionDef):
                    functions.append(class_node)
    return functions

def get_list_of_functions(filepath: str) -> List[Function]:
    tree = load_and_parse_file_for_tree(filepath)
    functions = []
    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            params = extract_parameter_types(node)
            start_line = node.lineno
            end_line = max(
                [line.lineno for line in ast.walk(node) if hasattr(line, 'lineno') and line.lineno],
                default=start_line
            )
            num_lines = end_line - start_line + 1
            name = node.name
            source_file_id = -1  # Placeholder
            functions.append(Function(name, str(params), start_line, end_line, num_lines, source_file_id))
        if isinstance(node, ast.ClassDef):
            for method in node.body:
                if isinstance(method, ast.FunctionDef):
                    params = extract_parameter_types(method)
                    start_line = method.lineno
                    end_line = max(
                        [line.lineno for line in ast.walk(method) if hasattr(line, 'lineno') and line.lineno],
                        default=start_line
                    )
                    num_lines = end_line - start_line + 1

                    # Method name (prefixed with class name for clarity)
                    name = f"{node.name}.{method.name}"

                    # Source file ID will be set when adding to database
                    source_file_id = -1  # Placeholder

                    functions.append(Function(name, str(params), start_line, end_line, num_lines, source_file_id))
    return functions

def get_function_boundaries(file_name: str, func_name: str) -> tuple:
    tree = load_and_parse_file_for_tree(file_name)
    for i, node in enumerate(tree.body):
        if isinstance(node, ast.FunctionDef) and node.name == func_name:
            func_start = node.lineno

            if i == len(tree.body) - 1:
                max_lines = [line.lineno for line in ast.walk(node) if hasattr(line, 'lineno') and line.lineno]
                func_end = max(max_lines) if max_lines else func_start
            else:
                next_node = tree.body[i + 1]
                if hasattr(next_node, 'lineno'):
                    func_end = next_node.lineno - 1
                else:
                    max_lines = [line.lineno for line in ast.walk(node) if hasattr(line, 'lineno') and line.lineno]
                    func_end = max(max_lines) if max_lines else func_start

            return func_start, func_end
        
    # For classes
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            for i, method in enumerate(node.body):
                if isinstance(method, ast.FunctionDef) and method.name == func_name:
                    func_start = method.lineno
                    
                    # Find end of method
                    if i == len(node.body) - 1:
                        max_lines = [line.lineno for line in ast.walk(method) 
                                     if hasattr(line, 'lineno') and line.lineno]
                        func_end = max(max_lines) if max_lines else func_start
                    else:
                        next_method = node.body[i + 1]
                        if hasattr(next_method, 'lineno'):
                            func_end = next_method.lineno - 1
                        else:
                            max_lines = [line.lineno for line in ast.walk(method)
                                         if hasattr(line, 'lineno') and line.lineno]
                            func_end = max(max_lines) if max_lines else func_start
                    
                    return func_start, func_end
    
    raise ValueError(f"Function {func_name} not found in {file_name}")

def parse_test_case_from_result_name(name: str, format: int) -> TestCase:
    test_case = TestCase(None, (), None)
    if format == 3:
        test_case = parse_test_case_from_result_name_doctest(name)
    elif format == 2:
        test_case = parse_test_case_from_result_name_pytest(name, format)
    elif format == 1:
        test_case = parse_test_case_from_result_name_unittest(name, format)
    else:
        raise ValueError(f"Unsupported format: {format}")
    return test_case

def parse_test_case_from_result_name_pytest(name: str, format: int) -> TestCase:
    parts_name = name.split("::")
    test_filepath = parts_name[0]
    print(f"Parse test case from result name pytest: {test_filepath}")
    test_filepath = os.path.abspath(os.path.join(os.getcwd(), test_filepath))
    test_function_name = parts_name[1]

    if test_function_name.startswith("test_"):
        function_name_with_int = test_function_name[5:]
    else:
        function_name_with_int = test_function_name

    parts = function_name_with_int.rsplit('_', 1)
    if len(parts) > 1 and parts[1].isdigit():
        function_name = parts[0]
    else:
        function_name = function_name_with_int

    args, expected = extract_test_case_data(test_filepath, test_function_name, format)

    return TestCase(function_name, args, expected)

def parse_test_case_from_result_name_unittest(name: str, format: int) -> TestCase:
    parts_name = name.split("::")
    test_filepath = parts_name[0]
    test_function_name_with_mod = parts_name[1]
    test_function_name = test_function_name_with_mod.split('.')[1]

    args, expected = extract_test_case_data(test_filepath, test_function_name, format)

    parts = test_function_name.rsplit('_', 1)
    if len(parts) > 1 and parts[1].isdigit():
        function_name = parts[0]
    else:
        function_name = test_function_name

    return TestCase(function_name, args, expected)

# TODO: Find a way to associate doctest results to the actual test case
def parse_test_case_from_result_name_doctest(name: str) -> TestCase:
    function_name = name.split("::")[1]
    return TestCase(function_name, (), None)

def extract_test_case_data(test_file_path: str, function_name: str, test_format: int):
    if not os.path.exists(test_file_path):
        print(f"Error: Test file not found: {test_file_path}")
        return {}, None

    try:
        with open(test_file_path, 'r') as f:
            file_content = f.read()

        # For doctests, just return empty data since we can't easily extract it
        if test_format == 3:  # Doctest format
            return {}, None

        # Parse the file using ast
        tree = ast.parse(file_content)

        # Look for test function definition based on format
        if test_format == 1:  # Unittest
            return _extract_unittest_data(tree, function_name)
        elif test_format == 2:  # Pytest
            return _extract_pytest_data(tree, function_name)
        else:
            return None

    except Exception as e:
        print(f"Error extracting test case data: {e}")
        return None

def _extract_unittest_data(tree: ast.AST, function_name: str):
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            # Look for test method within the class
            for class_node in node.body:
                if isinstance(class_node, ast.FunctionDef) and class_node.name == f"test_{function_name}":
                    # Find 'args =' and 'expected =' assignments in the function body
                    args_value = None
                    expected_value = None

                    for stmt in class_node.body:
                        if isinstance(stmt, ast.Assign):
                            if isinstance(stmt.targets[0], ast.Name):
                                if stmt.targets[0].id == "args":
                                    args_value = _extract_value(stmt.value)
                                elif stmt.targets[0].id == "expected":
                                    expected_value = _extract_value(stmt.value)

                    if args_value is not None:
                        # Convert args to dict for consistency
                        if isinstance(args_value, tuple):
                            return args_value, expected_value
                        else:
                            return (args_value,), expected_value

    return None

def _extract_pytest_data(tree: ast.AST, function_name: str):
    """Extract args and expected output from pytest-style test file"""
    for node in tree.body:
        if isinstance(node, ast.FunctionDef) and node.name == f"test_{function_name}":
            # Find 'args =' and 'expected =' assignments in the function body
            args_value = None
            expected_value = None

            for stmt in node.body:
                if isinstance(stmt, ast.Assign):
                    if isinstance(stmt.targets[0], ast.Name):
                        if stmt.targets[0].id == "args":
                            args_value = _extract_value(stmt.value)
                        elif stmt.targets[0].id == "expected":
                            expected_value = _extract_value(stmt.value)

            if args_value is not None:
                # Convert args to dict for consistency
                if isinstance(args_value, tuple):
                    return args_value, expected_value
                else:
                    return (args_value,), expected_value

    return None

def _extract_value(node: ast.AST):
    """Convert an AST node to its Python value"""
    if isinstance(node, ast.Tuple):
        return tuple(_extract_value(elt) for elt in node.elts)
    elif isinstance(node, ast.List):
        return [_extract_value(elt) for elt in node.elts]
    elif isinstance(node, ast.Dict):
        return {_extract_value(key): _extract_value(value) for key, value in zip(node.keys, node.values)}
    elif isinstance(node, ast.Constant):
        return node.value


    # If we can't determine the value, try using ast.unparse (Python 3.9+)
    try:
        return eval(ast.unparse(node))
    except:
        try:
            # Fallback for older Python versions
            code = compile(ast.Expression(node), '<string>', 'eval')
            return eval(code)
        except:
            return None





import ast
from typing import List, Any

import coverage

from testgen.models.analysis_context import AnalysisContext
from testgen.models.function_metadata import FunctionMetadata
from testgen.models.test_case import TestCase
from testgen.util.file_utils import load_and_parse_file_for_tree, load_module
from testgen.util.utils import get_function_boundaries
from testgen.util.z3_utils.constraint_extractor import extract_branch_conditions


def get_coverage_analysis(filepath: str, function_metadata: FunctionMetadata, args) -> tuple:
    func_node = function_metadata.func_def
    func_start = func_node.lineno
    func_end = max([node.lineno for node in ast.walk(func_node) if hasattr(node, 'lineno')], default=func_start)

    func = function_metadata.func

    analysis_list = _get_coverage_analysis_list(func, filepath, args)

    analysis_list[1] = [line for line in analysis_list[1] if func_start <= line <= func_end]
    analysis_list[3] = [line for line in analysis_list[3] if func_start <= line <= func_end]

    branch_lines = _get_branch_lines(func_node)

    for line in branch_lines:
        if func_start <= line <= func_end and line not in analysis_list[1]:
            analysis_list[1].append(line)
    analysis_list[1].sort()

    if func_start not in analysis_list[1]:
        analysis_list[1].append(func_start)
        analysis_list[1].sort()

    if func_start in analysis_list[3]:
        analysis_list[3].remove(func_start)

    return tuple(analysis_list)


def _get_coverage_analysis_list(func: Any, filepath: str, args) -> list:
    cov = coverage.Coverage(branch=True)
    cov.start()

    func(*args)

    cov.stop()
    cov.save()

    analysis = cov.analysis2(filepath)
    analysis_list = list(analysis)
    return analysis_list

def _get_branch_lines(func_node: ast.FunctionDef) -> list:
    branch_lines = []
    for node in ast.walk(func_node):
        if isinstance(node, ast.If):
            branch_lines.append(node.lineno)
            if node.orelse:
                for else_item in node.orelse:
                    if hasattr(else_item, 'lineno'):
                        else_line = else_item.lineno - 1
                        branch_lines.append(else_line)
                        break
    return branch_lines

def get_list_of_covered_statements(analysis: tuple) -> list:
    return [x for x in analysis[1] if x not in analysis[3]]

def get_uncovered_lines_for_func(file_name: str, class_name: str | None, func_node: ast.FunctionDef, test_cases: List[TestCase]) -> List[int]:
    # Get normal uncovered lines
    func_name = func_node.name

    function_test_cases = [tc for tc in test_cases if tc.func_name == func_name]
    if not function_test_cases:
        print(f"Warning: No test cases found for function {func_name}.")
        return []

    module = load_module(file_name)
    if class_name is not None:
        class_obj = getattr(module, class_name)
        instance = class_obj()
        func = getattr(instance, func_name)
    else:
        func = getattr(module, func_name)

    # Run coverage
    cov = coverage.Coverage(branch=True)  # Enable branch coverage
    cov.start()
    for test_case in function_test_cases:
        if test_case.func_name == func_name:
            try:
                func(*test_case.inputs)
            except Exception as e:
                print(f"Warning: Test Case {test_case.inputs} failed with error: {e}")
    cov.stop()

    analysis = cov.analysis2(file_name)

    # Extract branch conditions from the function
    branch_conditions, _ = extract_branch_conditions(func_node)
    condition_line_numbers = [bc.line_number for bc in branch_conditions]

    # Check which branch condition lines weren't exercised
    func_start, func_end = get_function_boundaries(file_name, func_name)
    missed_lines = [line for line in analysis[3] if func_start <= line <= func_end]

    # Find branch conditions that need to be tested (those near missed lines)
    uncovered_branch_lines = []
    for line in condition_line_numbers:
        # Check if the condition itself or its following line (likely the branch body) is uncovered
        if line in missed_lines or (line + 1) in missed_lines:
            uncovered_branch_lines.append(line)

    return uncovered_branch_lines

def get_all_executable_statements(filepath: str, func: FunctionMetadata, test_cases):
    import ast
    from testgen.util.randomizer import new_random_test_case

    if not test_cases:
        temp_case = new_random_test_case(filepath, func.class_name, func.func_def)
        analysis = get_coverage_analysis(filepath, func, temp_case.inputs)
    else:
        analysis = get_coverage_analysis(filepath, func, test_cases[0].inputs)

    executable_lines = list(analysis[1])

    with open(filepath, 'r') as f:
        source = f.read()

    tree = ast.parse(source)

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name == func.func_def.name:
            for if_node in ast.walk(node):
                if isinstance(if_node, ast.If) and if_node.orelse:
                    if isinstance(if_node.orelse[0], ast.If):
                        continue
                    else_line = if_node.orelse[0].lineno - 1

                    with open(filepath, 'r') as f:
                        lines = f.readlines()
                        if else_line <= len(lines):
                            line_content = lines[else_line - 1].strip()
                            if line_content == "else:":
                                if else_line not in executable_lines:
                                    executable_lines.append(else_line)

    return sorted(executable_lines)


from abc import ABC, abstractmethod
from typing import List, Tuple, Any

from testgen.models.function import Function


class Dao(ABC):
    @abstractmethod
    def insert_test_suite(self, name: str) -> int:
        pass

    @abstractmethod
    def insert_source_file(self, path: str, lines_of_code: int, last_modified) -> int:
        pass

    @abstractmethod
    def insert_function(self, name: str, params, start_line: int, end_line: int, source_file_id: int) -> int:
        pass

    @abstractmethod
    def insert_test_case(self, test_case: Any, test_suite_id: int, function_id: int, test_method_type: int) -> int:
        pass

    @abstractmethod
    def insert_test_result(self, test_case_id: int, status: bool, error: str = None) -> int:
        pass

    @abstractmethod
    def insert_coverage_data(self, file_name: str, executed_lines: int, missed_lines: int,
                             branch_coverage: float, source_file_id: int, function_id: int = None) -> int:
        pass

    @abstractmethod
    def get_test_suites(self) -> List[Any]:
        pass

    @abstractmethod
    def get_test_cases_by_function(self, function_name: str) -> List[Any]:
        pass

    @abstractmethod
    def get_source_file_id_by_path(self, filepath: str) -> int:
        pass

    @abstractmethod
    def get_coverage_by_file(self, file_path: str) -> List[Any]:
        pass

    @abstractmethod
    def get_test_file_data(self, file_path: str) -> List[Any]:
        pass

    @abstractmethod
    def get_function_by_name_file_id_start(self, name: str, source_file_id: int, start_line: int)-> int:
        pass

    @abstractmethod
    def get_functions_by_file(self, filepath: str) -> List[Function]:
        pass

    @abstractmethod
    def get_test_suite_id_by_name(self, name: str) -> int:
        pass

    @abstractmethod
    def get_test_case_id_by_func_id_input_expected(self, function_id: int, inputs: str, expected: str) -> int:
        pass




import random
import time
import traceback
from typing import List, Dict, Set
import testgen.util.coverage_utils as coverage_utils
from testgen.analyzer.contracts.contract import Contract
from testgen.analyzer.contracts.no_exception_contract import NoExceptionContract
from testgen.analyzer.contracts.nonnull_contract import NonNullContract
from testgen.models.test_case import TestCase
from testgen.analyzer.test_case_analyzer import TestCaseAnalyzerStrategy
from abc import ABC

from testgen.models.function_metadata import FunctionMetadata


# Citation in which this method and algorithm were taken from:
# C. Pacheco, S. K. Lahiri, M. D. Ernst and T. Ball, "Feedback-Directed Random Test Generation," 29th International
# Conference on Software Engineering (ICSE'07), Minneapolis, MN, USA, 2007, pp. 75-84, doi: 10.1109/ICSE.2007.37.
# keywords: {System testing;Contracts;Object oriented modeling;Law;Legal factors;Open source software;Software
# testing;Feedback;Filters;Error correction codes},

class RandomFeedbackAnalyzer(TestCaseAnalyzerStrategy, ABC):
    def __init__(self, analysis_context=None):
        super().__init__(analysis_context)
        self.test_cases = []
        self.covered_lines: Dict[str, Set[int]] = {}
        self.covered_functions: Set[str] = set()

    def collect_test_cases(self, function_metadata: FunctionMetadata, time_limit: int = 5) -> List[TestCase]:
        self.test_cases = []
        start_time = time.time()
        
        while (time.time() - start_time) < time_limit:
            
            try:
                param_values = self.generate_random_inputs(function_metadata.params)
                func_name = function_metadata.function_name
                function = function_metadata.func
                
                param_names = function_metadata.params.keys()
                
                ordered_args = [param_values.get(name, None) for name in param_names]
                
                result = function(*ordered_args)
                test_case = TestCase(func_name, tuple(ordered_args), result)

                if not self.is_duplicate_test_case(test_case):
                    self.test_cases.append(test_case)
                
                    covered = self.covered(function_metadata)
                    if covered:
                        break
                else:
                    # Optionally log duplicate detection
                    self.logger.debug(f"Skipping duplicate test case: {func_name}{test_case.inputs}")
                            
            except Exception as e:
                print(f"Error testing {function_metadata.function_name}: {e}")

        return self.test_cases
    
    def is_duplicate_test_case(self, new_test_case: TestCase) -> bool:
        for existing_test_case in self.test_cases:
            if (existing_test_case.func_name == new_test_case.func_name and
                existing_test_case.inputs == new_test_case.inputs):
                return True
        return False

    def covered(self, func: FunctionMetadata) -> bool:
        if func.function_name not in self.covered_lines:
            self.covered_lines[func.function_name] = set()

        test_cases = [tc for tc in self.test_cases if tc.func_name == func.function_name]

        for test_case in test_cases:
            analysis = coverage_utils.get_coverage_analysis(self._analysis_context.filepath, func, test_case.inputs)
            covered = coverage_utils.get_list_of_covered_statements(analysis)
            self.covered_lines[func.function_name].update(covered)
            self.logger.debug(f"Covered lines for {func.function_name}: {self.covered_lines[func.function_name]}")


        executable_statements = set(coverage_utils.get_all_executable_statements(self._analysis_context.filepath, func, test_cases))
        self.logger.debug(f"Executable statements for {func.function_name}: {executable_statements}")

        return self.covered_lines[func.function_name] == executable_statements

    def execute_sequence(self, sequence, contracts: List[Contract]):
        """Execute a sequence and check contract violations"""
        func_name, args_dict = sequence

        try:
            # Use module from analysis context if available
            function_metadata = self.get_function_metadata(func_name)
            func = function_metadata.func
            param_names = function_metadata.params.keys()

            ordered_args = [args_dict.get(name, None) for name in param_names]

            # Check preconditions
            for contract in contracts:
                if not contract.check_preconditions(tuple(ordered_args)):
                    print(f"Preconditions failed for {func_name} with {tuple(ordered_args)}")
                    return None, True

            # Execute function with properly ordered arguments
            output = func(*ordered_args)
            exception = None

        except Exception as e:
            print(f"EXCEPTION IN RANDOM FEEDBACK: {e}")
            print(traceback.format_exc())
            output = None
            exception = e

        # Check postconditions
        for contract in contracts:
            if not contract.check_postconditions(tuple(ordered_args), output, exception):
                print(f"Postcondition failed for {func_name} with {tuple(ordered_args)}")
                return output, True

        return output, False

    def get_function_metadata(self, func_name: str) -> FunctionMetadata | None:
        for function_data in self._analysis_context.function_data:
            if function_data.function_name == func_name:
                return function_data
        return None

    # TODO: Currently only getting random vals of primitives, extend to sequences
    def random_seqs_and_vals(self, param_types, non_error_seqs=None):
        return self.generate_random_inputs(param_types)

    @staticmethod
    def generate_random_inputs(param_types):
        """Generate inputs for fuzzing based on parameter types."""
        inputs = {}
        for param, param_type in param_types.items():
            if param_type == "int":
                random_integer = random.randint(-500, 500)  # Wider range for better edge cases
                inputs[param] = random_integer
            elif param_type == "bool":
                random_choice = random.choice([True, False])
                inputs[param] = random_choice
            elif param_type == "float":
                random_float = random.uniform(-500.0, 500.0)  # Wider range for better edge cases
                inputs[param] = random_float
            elif param_type == "str":
                # Generate diverse strings instead of always "abc"
                string_type = random.choice([
                    "empty", "short", "medium", "long", "special", "numeric", "whitespace"
                ])
                
                if string_type == "empty":
                    inputs[param] = ""
                elif string_type == "short":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyz', k=random.randint(1, 3)))
                elif string_type == "medium":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', k=random.randint(4, 10)))
                elif string_type == "long":
                    inputs[param] = ''.join(random.choices('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789', k=random.randint(11, 30)))
                elif string_type == "special":
                    inputs[param] = ''.join(random.choices('!@#$%^&*()_+-=[]{}|;:,./<>?', k=random.randint(1, 8)))
                elif string_type == "numeric":
                    inputs[param] = ''.join(random.choices('0123456789', k=random.randint(1, 10)))
                else:  # whitespace
                    inputs[param] = ' ' * random.randint(1, 5)
            else:
                # For unknown types, try a default value
                inputs[param] = None
                
        return inputs
    
    # Algorithm described in above article
    # Classes is the classes for which we want to generate sequences
    # Contracts express invariant properties that hold both at entry and exit from a call
        # Contract takes as input the current state of the system (runtime values created in the sequence so far, and any exception thrown by the last call), and returns satisfied or violated
        # Output is the runtime values and boolean flag violated
    # Filters determine which values of a sequence are extensible and should be used as inputs
    def generate_sequences(self, function_metadata: List[FunctionMetadata], classes=None, contracts: List[Contract] = None, filters=None, time_limit=20):
        contracts = [NonNullContract(), NoExceptionContract()]
        error_seqs = [] # execution violates a contract
        non_error_seqs = [] # execution does not violate a contract

        functions = self._analysis_context.function_data
        start_time = time.time()
        while(time.time() - start_time) >= time_limit:
            # Get random function
            func = random.choice(functions)
            param_types: dict = func.params
            vals: dict = self.random_seqs_and_vals(param_types)
            new_seq = (func.function_name, vals)
            if new_seq in error_seqs or new_seq in non_error_seqs:
                continue
            outs_violated: tuple = self.execute_sequence(new_seq, contracts)
            violated: bool = outs_violated[1]
            # Create tuple of sequence ((func name, args), output)
            new_seq_out = (new_seq, outs_violated[0])
            if violated:
                error_seqs.append(new_seq_out)
            else:
                # Question: Should I use the failed contract to be the assertion in unit test??
                non_error_seqs.append(new_seq_out)
        return error_seqs, non_error_seqs

    def generate_sequences_new(self, contracts: List[Contract] = None, filters=None, time_limit=20):
        contracts = [NonNullContract(), NoExceptionContract()]
        error_seqs = []  # execution violates a contract
        non_error_seqs = []  # execution does not violate a contract

        functions = self._analysis_context.function_data.copy()
        start_time = time.time()

        while (time.time() - start_time) < time_limit:
            # Get random function
            func = random.choice(functions)
            param_types: dict = func.params
            vals: dict = self.random_seqs_and_vals(param_types)
            new_seq = (func.function_name, vals)

            if new_seq in [seq[0] for seq in error_seqs] or new_seq in [seq[0] for seq in non_error_seqs]:
                continue

            outs_violated: tuple = self.execute_sequence(new_seq, contracts)
            violated: bool = outs_violated[1]

            # Create tuple of sequence ((func name, args), output)
            new_seq_out = (new_seq, outs_violated[0])

            if violated:
                error_seqs.append(new_seq_out)

            else:
                non_error_seqs.append(new_seq_out)

            test_case = TestCase(new_seq_out[0][0], tuple(new_seq_out[0][1].values()), new_seq_out[1])
            self.test_cases.append(test_case)
            fully_covered = self.covered(func)
            if fully_covered:
                print(f"Function {func.function_name} is fully covered")
                functions.remove(func)

            if not functions:
                self.test_cases.sort(key=lambda tc: tc.func_name)
                print("All functions covered")
                break

        self.test_cases.sort(key=lambda tc: tc.func_name)
        return error_seqs, non_error_seqs


import doctest
import unittest
import subprocess
import sys
import os
from typing import Dict, List, Any

import pytest
from _pytest.reports import TestReport

from testgen.service.logging_service import get_logger
from testgen.util import utils
from testgen.service.db_service import DBService


class TestExecutorService:
    UNITTEST_FORMAT = 1
    PYTEST_FORMAT = 2
    DOCTEST_FORMAT = 3

    def __init__(self):
        self.logger = get_logger()

    def execute_tests(self, test_file: str, test_format: int) -> List[Dict[str, Any]]:

        test_file = os.path.abspath(test_file)
        if not os.path.exists(test_file):
            self.logger.error(f"Test file not found: {test_file}")
            return [{"name": f"{test_file}::file_not_found", "status": False, "error": "Test file not found"}]

        try:
            if test_format == self.UNITTEST_FORMAT:
                return self.execute_unittest(test_file)
            elif test_format == self.PYTEST_FORMAT:
                return self.execute_pytest(test_file)
            elif test_format == self.DOCTEST_FORMAT:
                return self.execute_doctest(test_file)
            else:
                self.logger.error(f"Unsupported test format: {test_format}")
                return [{"name": f"{test_file}::invalid_format", "status": False,
                         "error": f"Unsupported test format: {test_format}"}]

        except Exception as e:
            self.logger.error(f"Error executing tests: {str(e)}")
            return [{"name": f"{test_file}::execution_error", "status": False, "error": str(e)}]

    # Currently not collecting results
    def execute_unittest(self, test_file: str) -> List[Dict[str, Any]]:
        print(f"Running unittest on: {test_file}")
        result = subprocess.run(
            [sys.executable, "-m", "unittest", test_file],
            capture_output=True,
            text=True
        )
        print(result.stdout)
        print(result.stderr)
        # Would be used to return results
        return []

    def execute_pytest(self, test_file: str) -> List[Dict[str, Any]]:
        print(f"Executing pytest file: {test_file}")
        results = []

        try:
            # Custom plugin to collect results
            class PytestResultCollector:
                def __init__(self):
                    self.results = []

                def pytest_runtest_logreport(self, report: TestReport):
                    if report.when == "call":
                        self.results.append({
                            "name": report.nodeid,
                            "status": report.outcome == "passed",
                            "error": report.longreprtext if report.outcome != "passed" else None
                        })

            # Run pytest and collect results
            collector = PytestResultCollector()
            pytest.main([test_file], plugins=[collector])
            results = collector.results

        except Exception as e:
            self.logger.error(f"Error running pytest: {e}")
            results.append({
                "name": f"{test_file}::pytest_execution",
                "status": False,
                "error": str(e)
            })

        return results

    # Currently not collecting results
    def execute_doctest(self, test_file: str) -> List[Dict[str, Any]]:
        print(f"Running doctest on: {test_file}")
        #failed, attempted = doctest.testfile(test_file, module_relative=False, verbose=False)
        # doctest prints results by itself when verbose=True

        # Would be used to collect and return results
        return []

    # Currently not in use and not working to save the test results since I am having difficulty
    # getting the test case ID for each method and ran out of time :(
    def save_test_results(self, db_service: DBService, test_results: List[Dict[str, Any]],
                          file_path: str, test_format: int) -> None:
        if db_service is None:
            self.logger.debug("Skipping database operations - no DB service provided")
            return

        try:
            source_file_id = db_service.get_source_file_id_by_path(file_path)

            if source_file_id == -1:
                self.logger.error(f"Source file not found in database: {file_path}")
                return

            functions = db_service.get_functions_by_file(file_path)

            for result in test_results:
                name = result["name"]
                test_case = utils.parse_test_case_from_result_name(name, test_format)
                print(f"SAVE TEST RESULTS TEST CASE {test_case}")
                function_id = db_service.match_test_case_to_function_for_id(source_file_id, test_case, functions)
                print(f"SAVE TEST RESULTS FUNCTION ID {function_id}")

                if function_id == -1:
                    self.logger.warning(f"Could not match test case {name} to a function")
                    continue

                inputs_str = str(test_case.inputs)
                print(f"Inputs str {inputs_str}")
                expected_str = str(test_case.expected)
                print(f"Expected str {expected_str}")

                test_case_id = db_service.get_test_case_id_by_func_id_input_expected(
                    function_id, inputs_str, expected_str)

                if test_case_id == -1:
                    self.logger.warning(f"Test case not found in database: {name}")
                    continue

                db_service.insert_test_result(test_case_id, result["status"], result["error"])

        except Exception as e:
            self.logger.error(f"Error saving test results to database: {e}")


import os
import inspect
from types import ModuleType
from typing import List

from testgen.models.test_case import TestCase
from testgen.generator.code_generator import CodeGenerator
from testgen.generator.doctest_generator import DocTestGenerator
from testgen.generator.pytest_generator import PyTestGenerator
from testgen.generator.unit_test_generator import UnitTestGenerator
from testgen.service.logging_service import get_logger
from testgen.tree.node import Node
from testgen.tree.tree_utils import build_binary_tree
from testgen.models.generator_context import GeneratorContext
from testgen.util.file_utils import find_project_root

# Constants for test formats
UNITTEST_FORMAT = 1
PYTEST_FORMAT = 2
DOCTEST_FORMAT = 3

class GeneratorService:
    def __init__(self, filepath: str, output_path: str, test_format: int = UNITTEST_FORMAT):
        self.filepath = filepath
        self.output_path = output_path
        self.test_format = test_format
        self.code_generator = CodeGenerator()
        self.test_generator = UnitTestGenerator(generator_context=None)
        self.generated_file_path = None
        self.logger = get_logger()

    def set_test_format(self, test_format: int):
        """Set the test generator format."""
        self.test_format = test_format
        if test_format == UNITTEST_FORMAT:
            self.logger.debug("SETTING TEST FORMAT TO UNITTEST")
            self.test_generator = UnitTestGenerator(generator_context=None)
        elif test_format == PYTEST_FORMAT:
            self.logger.debug("SETTING TEST FORMAT TO PYTEST")
            self.test_generator = PyTestGenerator(generator_context=None)
        elif test_format == DOCTEST_FORMAT:
            self.logger.debug("SETTING TEST FORMAT TO DOCTEST")
            self.test_generator = DocTestGenerator(generator_context=None)
        else:
            raise NotImplementedError(f"Test format {test_format} not implemented")

    def generate_test_file(self, module: ModuleType, class_name: str | None, test_cases: List[TestCase], output_path=None) -> str:
        """Generate a test file for the given test cases."""
        filename = self.get_filename(self.filepath)
        output_path = self.get_test_file_path(module.__name__, output_path)
        
        actual_class_name = self.resolve_class_name(module, class_name)
        
        context = self.get_generator_context(self.filepath, module, actual_class_name, test_cases, output_path)

        self.test_generator.generator_context = context
        
        self.test_generator.generate_test_header()

        self.generate_function_tests(test_cases)

        if self.test_format == DOCTEST_FORMAT:
            self.logger.debug("SAVING DOCT TEST FILE")
            self.test_generator.save_file()
            return self.filepath
        else:
            self.test_generator.save_file()
            return output_path

    def generate_function_tests(self, test_cases: List[TestCase]) -> None:
        """Generate test functions for the given test cases."""
        for i, test_case in enumerate(test_cases):
            unique_func_name = f"{test_case.func_name}_{i}"
            cases = [(test_case.inputs, test_case.expected)]
            self.test_generator.generate_test_function(unique_func_name, test_case.func_name, cases)

    def get_test_file_path(self, module_name: str, specified_path=None) -> str:
        """Determine the path for the generated test file."""
        if specified_path is not None:
            if os.path.exists(specified_path):
                if os.path.isdir(specified_path):
                    self.ensure_init_py(specified_path)
                    return os.path.join(specified_path, f"test_{module_name.lower()}.py")
                else:
                    print(f"Specified directory path: {specified_path} is not a directory.")
            else:
                print(f"Specified directory path: {specified_path} does not exist.")

        current_dir = os.getcwd()
        test_dir = os.path.join(current_dir, "tests")

        if os.path.exists(test_dir):
            if os.path.isdir(test_dir):
                self.ensure_init_py(test_dir)
                return os.path.join(test_dir, f"test_{module_name.lower()}.py")
            else:
                print(f"Test directory path: {test_dir} is not a directory.")
        else:
            print(f"Test directory path: {test_dir} does not exist, creating it.")
            os.mkdir(test_dir)
            self.ensure_init_py(test_dir)

        return os.path.join(test_dir, f"test_{module_name.lower()}.py")

    def ensure_init_py(self, directory: str):
        """Ensures that an __init__.py file exists in the given directory."""
        init_file = os.path.join(directory, "__init__.py")
        if not os.path.exists(init_file):
            with open(init_file, "w") as f:
                pass  # Create an empty __init__.py file
            print(f"Created __init__.py in {directory}")

    def generate_function_code(self, file_path: str, class_name: str | None, functions: list) -> str:
        """Generate function code for a given class and its functions."""
        trees = self.build_func_trees(functions)
        if class_name:
            print(class_name)
            self.generated_file_path = f"generated_{class_name.lower()}.py"
        else:
            print(self.get_filename(file_path))
            self.generated_file_path = f"generated_{self.get_filename(file_path)}"

        # Create the class file
        if class_name:
            file = self.code_generator.generate_class(class_name)
            file.close()
        else:
            file = open(self.generated_file_path, "w")
            file.close()

        # Append function implementations
        with open(self.generated_file_path, "a") as file:
            for func, root, params in trees:
                is_class_method = True if class_name else False
                
                # Get parameter types
                param_types = {}
                sig = inspect.signature(func)
                for param_name, param in sig.parameters.items():
                    if param_name != 'self':  # Skip self for class methods
                        # Extract type annotation if available
                        if param.annotation != inspect.Parameter.empty:
                            param_types[param_name] = param.annotation.__name__
                        else:
                            param_types[param_name] = 'object'  # Default to object if no annotation
                
                code = self.code_generator.generate_code_from_tree(
                    func.__name__, root, params, func, is_class_method, param_types
                )
                
                if class_name:
                    for line in code.split("\n"):
                        file.write(f"    {line}\n")
                else:
                    file.write(code)
                file.write("\n")

        return self.generated_file_path

    def get_generator_context(self, filepath: str, module: ModuleType, class_name: str | None, test_cases: List[TestCase], output_path: str) -> GeneratorContext:
        file_dir = os.path.dirname(os.path.abspath(filepath))
        is_package = os.path.exists(os.path.join(file_dir, '__init__.py'))

        package_name, import_path = self._resolve_package_name_and_import_path(filepath)

        generator_context = GeneratorContext(
            filepath=filepath,
            filename=self.get_filename(filepath),
            class_name=class_name,
            module=module,
            output_path=output_path,
            test_cases=test_cases,
            is_package=is_package,
            package_name=package_name,
            import_path=import_path
        )

        print(f"Generator Context: {generator_context.filepath} {generator_context.filename} {generator_context.class_name} {generator_context.is_package} {generator_context.package_name} {generator_context.import_path}")

        return generator_context

    @staticmethod
    def _resolve_package_name_and_import_path(filepath: str) -> tuple:
        if not os.path.exists(filepath) or not filepath.endswith('.py'):
            raise ValueError(f"Invalid Python file: {filepath}")

        # Get the directory and filename
        file_dir = os.path.dirname(os.path.abspath(filepath))
        module_name = os.path.splitext(os.path.basename(filepath))[0]

        # Find the project root by looking for setup.py or a .git directory
        project_root = find_project_root(file_dir)

        # Build the import path based on the file's location relative to the project root
        if project_root:
            rel_path = os.path.relpath(file_dir, project_root)
            if rel_path == '.':
                # File is directly in the project root
                import_path = module_name
                package_name = ''
            else:
                # File is in a subdirectory
                path_parts = rel_path.replace('\\', '/').split('/')
                # Filter out any empty parts
                path_parts = [part for part in path_parts if part]

                if path_parts:
                    package_name = path_parts[0]
                    # Construct the full import path
                    import_path = '.'.join(path_parts) + '.' + module_name
                else:
                    package_name = ''
                    import_path = module_name
        else:
            # Fallback if we can't find a project root
            package_name = ''
            import_path = module_name

        return package_name, import_path

    def build_func_trees(self, functions: list):
        """Build binary trees for function signatures."""
        tree_list = []
        for name, func in functions:
            signature = inspect.signature(func)
            params = [param for param, value in signature.parameters.items() if param != 'self']
            root = Node(None)
            build_binary_tree(root, 0, len(params))
            tree_list.append((func, root, params))
        return tree_list

    def resolve_class_name(self, module: ModuleType, class_name: str):
        # Determine the actual class name used in the module
        actual_class_name = class_name
        if 'generated_' in self.filepath and class_name:
            # For generated classes, find the actual class name in the module
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj):
                    actual_class_name = name
                    break
        return actual_class_name

    @staticmethod
    def get_filename(filepath: str) -> str:
        """Get filename from filepath."""
        return os.path.basename(filepath)


import ast
import importlib.util
import os
from _ast import Module
from types import ModuleType

def find_project_root(start_dir: str) -> str | None:
    current_dir = start_dir
    
    # Walk up the directory tree
    while current_dir:
        # Check for common project root indicators
        if (os.path.exists(os.path.join(current_dir, 'setup.py')) or
            os.path.exists(os.path.join(current_dir, '.git')) or
            os.path.exists(os.path.join(current_dir, 'pyproject.toml'))):
            return current_dir
            
        if os.path.exists(os.path.join(current_dir, '__main__.py')):
            return current_dir
        
        # Check for 'testgen' directory which is your project name
        if os.path.basename(current_dir) == 'testgen':
            parent_dir = os.path.dirname(current_dir)
            return parent_dir
            
        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:  # Reached the root
            break
        current_dir = parent_dir
    
    return None


def load_module(file_path: str) -> ModuleType:
    # Load a Python module from a file path.
    if file_path is None:
        raise ValueError("File path not set! Use set_file_path() to specify the path of the file")

    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_filename(filepath: str) -> str:
    """Get filename from filepath."""
    return os.path.basename(filepath)


def load_and_parse_file_for_tree(file) -> Module:
    with open(file) as f:
        code = f.read()
    tree = ast.parse(code)
    return tree

def adjust_file_path_for_docker(file_path: str) -> str:
    """Adjust the file path to be valid inside the Docker container."""
    print(f"Docker - adjusting path: {file_path}")

    # If already absolute to /controller, return as-is
    if file_path.startswith("/controller/"):
        print(f"Docker - already adjusted: {file_path}")
        return file_path

    # If relative, make absolute
    adjusted_path = f"/controller/{file_path.lstrip('/')}"
    print(f"Docker - adjusted to: {adjusted_path}")
    return adjusted_path

def get_project_root_in_docker(script_path) -> str:
    # Try to find project root by looking for pyproject.toml or similar
    current_dir = os.path.dirname(os.path.abspath(script_path))
    while current_dir != '/':
        if os.path.exists(os.path.join(current_dir, 'pyproject.toml')) or \
           os.path.exists(os.path.join(current_dir, 'setup.py')):
            project_root = current_dir
            break
        current_dir = os.path.dirname(current_dir)
    else:
        # Fallback - use parent of script dir
        project_root = os.path.dirname(os.path.dirname(os.path.abspath(script_path)))
    
    print(f"Project root directory: {project_root}")
    return project_root


class TestSuite:
    def __init__(self, name: str, creation_date):
        self._name = name
        self._creation_date = creation_date

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def creation_date(self):
        return self._creation_date

    @creation_date.setter
    def creation_date(self, value) -> None:
        self._creation_date = value


import os
import sys
from argparse import Namespace
import docker
from docker import DockerClient, client
from docker import errors
from docker.models.containers import Container
import importlib.resources as pkg_resources
import tempfile
import shutil

from testgen.service.logging_service import get_logger
from testgen.service.service import Service

AST_STRAT = 1
FUZZ_STRAT = 2
RANDOM_STRAT = 3

UNITTEST_FORMAT = 1
PYTEST_FORMAT = 2
DOCTEST_FORMAT = 3

class DockerController:
    def __init__(self):
        self.service = Service()
        self.debug_mode = False
        self.args = None
        self.logger = get_logger()

    def run_in_docker(self, project_root: str, docker_client: DockerClient, args: Namespace) -> bool:
        self.args = args
        self.debug_mode = True if args.debug else False
    
        os.environ["RUNNING_IN_DOCKER"] = "1"

        # Check if Docker image exists, build it if not
        image_name = "testgen-runner"
        self.get_image(docker_client, image_name, project_root)
        if not self.args.safe:
            self.logger.info("Docker image not found. Running locally...")
            return False

        src_path = os.path.abspath(args.file_path)
        dest_path = os.path.join(project_root, os.path.basename(src_path))
        try:
            shutil.copy(src_path, dest_path)
            print(f"Copied file from {src_path} to {dest_path}")
            self.logger.debug(f"Copied file from {src_path} to {dest_path}")
        except Exception as e:
            print(f"Failed to copy file: {e}")
            self.logger.error(f"Failed to copy file: {e}")
            sys.exit(1)

        docker_args = [os.path.basename(args.file_path)] + [arg for arg in sys.argv[2:] if arg != "--safe"]
        docker_args[0] = f"/controller/{docker_args[0]}"

        # Run the container with the same arguments
        try:
            self.debug(f"project_root: {project_root}")
            container = self.run_container(docker_client, image_name, docker_args, project_root)

            # Stream the logs to the console
            logs_output = self.get_logs(container)

            self.clean_up(dest_path)
            self.debug(logs_output)

        except Exception as e:
            print(f"Error running container: {e}")
            sys.exit(1)

        # Create the target directory if it doesn't exist
        if args.output is None:
            target_path = os.path.join(os.getcwd(), 'tests')
        else:
            target_path = args.output
        os.makedirs(target_path, exist_ok=True)

        self.debug(f"SERVICE target path after logs: {target_path}")

        test_cases = self.service.parse_test_cases_from_logs(logs_output)

        print(f"Extracted {len(test_cases)} test cases from container.")

        file_path = os.path.abspath(args.file_path)
        self.debug(f"Filepath in CLI CONTROLLER: {file_path}")
        self.service.set_file_path(file_path)

        if args.test_format == "pytest":
            self.service.set_test_generator_format(PYTEST_FORMAT)
        elif args.test_format == "doctest":
            self.service.set_test_generator_format(DOCTEST_FORMAT)
        else:
            self.service.set_test_generator_format(UNITTEST_FORMAT)

        test_file = self.service.generate_test_file(test_cases, target_path)
        print(f"Unit tests saved to: {test_file}")

        if not args.generate_only:
            print("Running coverage...")
            self.service.run_tests(test_file)
            self.service.run_coverage(test_file)
            
        # Add explicit return True here
        return True


    def get_image(self, docker_client: DockerClient, image_name: str, project_root: str):
        try:
            docker_client.images.get(image_name)
            print(f"Using existing {image_name} Docker image")
        except docker.errors.ImageNotFound:
            print(f"Building {image_name} Docker image...")

            # Look for Dockerfile in the project root
            dockerfile_path = self.get_dockerfile_path()
            if not os.path.exists(dockerfile_path):
                print(f"Dockerfile not found at {dockerfile_path}")
                sys.exit(1)

            self.debug(f"Using Dockerfile at: {dockerfile_path}")

            if not self.build_docker_image(docker_client, image_name, dockerfile_path, project_root):
                print("Failed to build Docker image. Continuing without safe mode.")
                self.args.safe = False

    @staticmethod
    def get_logs(container) -> str:
        # Stream the logs to the console
        logs = container.logs(stream=True)
        logs_output = ""
        for log in logs:
            log_line = log.decode()
            logs_output += log_line
            print(log_line, end="")
        return logs_output

    @staticmethod
    def run_container(docker_client: DockerClient, image_name: str, docker_args: list, project_root: str) -> Container:
        # Create Docker-specific environment variables
        docker_env = {
            "RUNNING_IN_DOCKER": "1",
            "PYTHONPATH": "/controller"
        }
        
        # Join arguments with proper escaping
        args_str = ' '.join(f'"{arg}"' for arg in docker_args)
        
        print(f"Docker args: {docker_args}")
        print(f"Project root: {project_root}")
        
        return docker_client.containers.run(
                image=image_name,
                command=["testgenie"] + docker_args,
                volumes={
                os.path.abspath(project_root): {
                "bind": "/controller",
                "mode": "rw"
                    }
                },
                working_dir="/controller",
                environment=docker_env,
                detach=True,
                remove=True,
                stdout=True,
                stderr=True
            )

    def build_docker_image(self, docker_client, image_name, dockerfile_path, project_root):
        try:
            print(f"Starting Docker build for image: {image_name}")
            dockerfile_rel_path = os.path.relpath(dockerfile_path, project_root)
            docker_dir = os.path.dirname(dockerfile_path)

            dockerfile_rel_path = os.path.relpath(dockerfile_path, project_root)
            self.debug(f"Project root {project_root}")
            self.debug(f"Docker directory: {docker_dir}")
            self.debug(f"Docker rel path: {dockerfile_rel_path}")

            build_progress = docker_client.api.build(
                path=docker_dir,          # now it knows what docker_dir is
                dockerfile=dockerfile_path,
                tag=image_name,
                rm=True,
                decode=True
            )

            for chunk in build_progress:
                self.debug(f"CHUNK: {chunk}")
                if 'stream' in chunk:
                    for line in chunk['stream'].splitlines():
                        if line.strip():
                            print(f"Docker: {line.strip()}")
                elif 'error' in chunk:
                    self.debug(f"Docker build error: {chunk['error']}")
                    return False
            print(f"Docker image built successfully: {image_name}")
            return True

        except docker.errors.BuildError as e:
            print(f"Docker build error: {e}")
        except docker.errors.APIError as e:
            print(f"Docker API error: {e}")
        except Exception as e:
            print(f"Unexpected error during Docker build: {str(e)}")
            return False

    def get_dockerfile_path(self) -> str:
        # First, try local development path
        local_dockerfile = os.path.join(os.path.dirname(__file__), "docker", "Dockerfile")
        if os.path.exists(local_dockerfile):
            self.debug(f"Found local Dockerfile at: {local_dockerfile}")
            return local_dockerfile

        # If not found locally, try inside installed package
        try:
            dockerfile_resource = pkg_resources.files('testgen').joinpath('docker/Dockerfile')
            if dockerfile_resource.is_file():
                self.debug(f"Found package-installed Dockerfile at: {dockerfile_resource}")
                return str(dockerfile_resource)
        except Exception as e:
            print(f"Error locating Dockerfile in package resources: {e}")

        print("Dockerfile not found in local project or package.")
        sys.exit(1)

    def clean_up(self, dest_path: str) -> None:
        # Remove copied file from project root
        try:
            if os.path.exists(dest_path):
                os.remove(dest_path)
                print(f"Deleted file from {dest_path}")
                self.logger.debug(f"Deleted file from {dest_path}")
        except Exception as e:
            print(f"Failed to delete file: {e}")
            self.logger.error(f"Failed to delete file: {e}")

    @staticmethod
    def is_inside_docker() -> bool:
        """Check if the current process is running inside a Docker container."""
        return os.environ.get("RUNNING_IN_DOCKER") in ("1", "true", "True")

    def debug(self, message: str):
        """Log debug message"""
        if self.debug_mode:
            self.logger.debug(message)


import os
import sqlite3
import json
from typing import List, Tuple, Any
from datetime import datetime

from testgen.db.dao import Dao
from testgen.models.function import Function
from testgen.models.test_case import TestCase
from testgen.db.db import create_database

class DaoImpl(Dao):
    def __init__(self, db_name="testgen.db"):
        self.db_name = db_name
        self.conn = None
        self.cursor = None
        self._connect()

    def _connect(self):
        if not os.path.exists(self.db_name):
            create_database(self.db_name)
        self.conn = sqlite3.connect(self.db_name)
        self.conn.row_factory = sqlite3.Row
        self.cursor = self.conn.cursor()
        self.cursor.execute("PRAGMA foreign_keys = ON;")

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None
            self.cursor = None

    def insert_test_suite(self, name: str) -> int:
        self.cursor.execute(
            "INSERT INTO TestSuite (name, creation_date) VALUES (?, ?)",
            (name, datetime.now())
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_source_file(self, path: str, lines_of_code: int, last_modified) -> int:
        self.cursor.execute("SELECT id FROM SourceFile WHERE path = ?", (path,))
        existing = self.cursor.fetchone()
        if existing:
            return existing[0]
        self.cursor.execute(
            "INSERT INTO SourceFile (path, lines_of_code, last_modified) VALUES (?, ?, ?)",
            (path, lines_of_code, last_modified)
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_function(self, name: str, params, start_line: int, end_line: int, source_file_id: int) -> int:
        print(f"INSERTING FUNCTION: {name}, {params}, {start_line}, {end_line}, {source_file_id}")

        num_lines = end_line - start_line + 1
        self.cursor.execute(
            "SELECT id FROM Function WHERE name = ? AND source_file_id = ? AND params = ?",
            (name, source_file_id, params)
        )
        existing = self.cursor.fetchone()
        if existing:
            return existing[0]
        self.cursor.execute(
            "INSERT INTO Function (name, params, start_line, end_line, num_lines, source_file_id) VALUES (?, ?, ?, ?, ?, ?)",
            (name, params, start_line, end_line, num_lines, source_file_id)
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_test_case(self, test_case: TestCase, test_suite_id: int, function_id: int, test_method_type: int) -> int:
        inputs_json = json.dumps(test_case.inputs, sort_keys=True)
        expected_json = json.dumps(test_case.expected, sort_keys=True)

        print(f"INSERTING TEST CASE: {test_case.func_name}, {test_case.inputs}, {test_case.expected}, {test_suite_id}, {function_id}, {test_method_type}")

        # Check for existing test case with same function_id, inputs, and expected_output
        self.cursor.execute(
            "SELECT id FROM TestCase WHERE function_id = ? AND input = ? AND expected_output = ?",
            (function_id, inputs_json, expected_json)
        )
        existing = self.cursor.fetchone()
        if existing:
            return existing[0]

        self.cursor.execute(
            "INSERT INTO TestCase (expected_output, input, test_function, last_run_time, test_method_type, test_suite_id, function_id) "
            "VALUES (?, ?, ?, ?, ?, ?, ?)",
            (
                expected_json,
                inputs_json,
                test_case.func_name,
                datetime.now(),
                test_method_type,
                test_suite_id,
                function_id
            )
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_test_result(self, test_case_id: int, status: bool, error: str = None) -> int:
        self.cursor.execute(
            "INSERT INTO TestResult (test_case_id, status, error, execution_time) VALUES (?, ?, ?, ?)",
            (test_case_id, status, error, datetime.now())
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def insert_coverage_data(self, file_name: str, executed_lines: int, missed_lines: int,
                             branch_coverage: float, source_file_id: int, function_id: int | None) -> int:
        coverage_type = "file" if function_id is None else "function"
        self.cursor.execute(
            "INSERT INTO CoverageData (coverage_type, executed_lines, missed_lines, branch_coverage, source_file_id, function_id) "
            "VALUES (?, ?, ?, ?, ?, ?)",
            (coverage_type, executed_lines, missed_lines, branch_coverage, source_file_id, function_id)
        )
        self.conn.commit()
        return self.cursor.lastrowid

    def get_test_suites(self) -> List[Any]:
        self.cursor.execute("SELECT * FROM TestSuite ORDER BY creation_date DESC")
        return self.cursor.fetchall()

    def get_test_cases_by_function(self, function_name: str) -> List[Any]:
        self.cursor.execute(
            "SELECT tc.* FROM TestCase tc JOIN Function f ON tc.function_id = f.id WHERE f.name = ?",
            (function_name,)
        )
        return self.cursor.fetchall()

    def get_coverage_by_file(self, file_path: str) -> List[Any]:
        self.cursor.execute(
            "SELECT cd.* FROM CoverageData cd JOIN SourceFile sf ON cd.source_file_id = sf.id WHERE sf.path = ?",
            (file_path,)
        )
        return self.cursor.fetchall()

    def get_test_file_data(self, file_path: str) -> List[Any]:
        query = """
            SELECT
                tc.id AS test_case_id,
                tc.name AS test_case_name,
                tc.test_function AS test_case_test_function,
                tc.test_method_type AS test_case_method_type,
                COALESCE(cd.missed_lines, 'None') AS coverage_data_missed_lines
            FROM SourceFile sf
            LEFT JOIN Function f ON sf.id = f.source_file_id
            LEFT JOIN TestCase tc ON f.id = tc.function_id
            LEFT JOIN TestResult tr ON tc.id = tr.test_case_id
            LEFT JOIN CoverageData cd ON sf.id = cd.source_file_id
            WHERE sf.path = ?;
        """
        self.cursor.execute(query, (file_path,))
        return self.cursor.fetchall()

    def get_function_by_name_file_id_start(self, name: str, source_file_id: int, start_line: int) -> int:
        print(f"GETTING FUNCTION ID FOR {name}, {source_file_id}, {start_line}")
        self.cursor.execute(
            """
            SELECT id FROM Function
            WHERE name = ? AND source_file_id = ? AND start_line = ?
            """,
            (name, source_file_id, start_line)
        )
        result = self.cursor.fetchone()
        return result["id"] if result else -1

    def get_functions_by_file(self, filepath: str) -> List[Function]:
        self.cursor.execute(
            """
            SELECT id, name, params, start_line, end_line, num_lines, source_file_id
            FROM Function
            WHERE source_file_id = (
                SELECT id FROM SourceFile WHERE path = ?
            )
            """,
            (filepath,)
        )
        rows = self.cursor.fetchall()
        return [
            Function(
                name=row["name"],
                params=row["params"],
                start_line=row["start_line"],
                end_line=row["end_line"],
                num_lines=row["num_lines"],
                source_file_id=row["source_file_id"]
            )
            for row in rows
        ]

    def get_test_suite_id_by_name(self, name: str) -> int:
        self.cursor.execute(
            """
            SELECT id FROM TestSuite
            WHERE name = ?
            """,
            (name,)
        )
        result = self.cursor.fetchone()
        return result["id"] if result else -1

    def get_source_file_id_by_path(self, filepath: str) -> int:
        self.cursor.execute(
            """
            SELECT id FROM SourceFile
            WHERE path = ?
            """,
            (filepath,)
        )
        result = self.cursor.fetchone()
        return result["id"] if result else -1

    def get_test_case_id_by_func_id_input_expected(self, function_id: int, inputs: str, expected: str) -> int:
        self.cursor.execute(
            """
            SELECT id FROM TestCase
            WHERE function_id = ? 
            AND input = ? 
            AND expected_output = ?
            """,
            (function_id, inputs, expected)
        )
        result = self.cursor.fetchone()
        return result["id"] if result else -1


from types import ModuleType
from typing import List
from testgen.models.test_case import TestCase


class GeneratorContext:
    def __init__(self, filepath: str, filename: str, class_name:str | None, module: ModuleType, output_path: str,
                 test_cases: List[TestCase], is_package: bool, package_name: str, import_path: str):
        self._filepath: str = filepath
        self._filename: str = filename
        self._class_name: str = class_name
        self._module: ModuleType = module
        self._output_path: str = output_path
        self._test_cases: List[TestCase] = test_cases
        self._is_package: bool = is_package
        self._package_name: str = package_name
        self._import_path: str = import_path

    @property
    def filepath(self) -> str:
        return self._filepath
    
    @filepath.setter
    def filepath(self, value: str) -> None:
        self._filepath = value
    
    @property
    def filename(self) -> str:
        return self._filename
    
    @filename.setter
    def filename(self, value: str) -> None:
        self._filename = value
    
    @property
    def class_name(self) -> str | None:
        return self._class_name
    
    @class_name.setter
    def class_name(self, value: str | None) -> None:
        self._class_name = value
    
    @property
    def module(self) -> ModuleType:
        return self._module
    
    @module.setter
    def module(self, value: ModuleType) -> None:
        self._module = value
    
    @property
    def output_path(self) -> str:
        return self._output_path
    
    @output_path.setter
    def output_path(self, value: str) -> None:
        self._output_path = value
    
    @property
    def test_cases(self) -> List[TestCase]:
        return self._test_cases
    
    @test_cases.setter
    def test_cases(self, value: List[TestCase]) -> None:
        self._test_cases = value

    @property
    def is_package(self) -> bool:
        return self._is_package

    @is_package.setter
    def is_package(self, value: bool) -> None:
        self._is_package = value

    @property
    def package_name(self) -> str:
        return self._package_name

    @package_name.setter
    def package_name(self, value: str) -> None:
        self._package_name = value

    @property
    def import_path(self) -> str:
        return self._import_path

    @import_path.setter
    def import_path(self, value: str) -> None:
        self._import_path = value




from abc import ABC, abstractmethod
from testgen.models.generator_context import GeneratorContext

class TestGenerator(ABC):
    def __init__(self, generator_context: GeneratorContext):
        self._generator_context = generator_context

    @property
    def generator_context(self) -> GeneratorContext:
        return self._generator_context

    @generator_context.setter
    def generator_context(self, value: GeneratorContext):
        self._generator_context = value

    @abstractmethod
    def generate_test_header(self):
        pass

    @abstractmethod
    def generate_test_function(self, unique_func_name, func_name, cases):
        pass

    @abstractmethod
    def save_file(self):
        pass